@@ -29,14 +29,22 @@
 #include "gibber-muc-connection.h"
 #include "gibber-xmpp-stanza.h"
 #include "gibber-namespaces.h"
+#include "gibber-xmpp-error.h"
 
 #define DEBUG_FLAG DEBUG_BYTESTREAM
 #include "gibber-debug.h"
 
 #include "signals-marshal.h"
-//#include "bytestream-factory.h"
 
-G_DEFINE_TYPE (GibberBytestreamIBB, gibber_bytestream_ibb, G_TYPE_OBJECT);
+/* FIXME: implement fragmentation using the "block-size" attribute */
+
+static void
+bytestream_iface_init (gpointer g_iface, gpointer iface_data);
+
+G_DEFINE_TYPE_WITH_CODE (GibberBytestreamIBB, gibber_bytestream_ibb,
+    G_TYPE_OBJECT,
+    G_IMPLEMENT_INTERFACE (GIBBER_TYPE_BYTESTREAM_IFACE,
+      bytestream_iface_init));
 
 /* signals */
 enum
@@ -52,28 +60,29 @@ static guint signals[LAST_SIGNAL] = {0};
 enum
 {
   PROP_XMPP_CONNECTION = 1,
-  PROP_MUC_CONNECTION,
   PROP_SELF_ID,
   PROP_PEER_ID,
   PROP_STREAM_ID,
   PROP_STREAM_INIT_ID,
   PROP_STATE,
+  PROP_PROTOCOL,
   LAST_PROPERTY
 };
 
 typedef struct _GibberBytestreamIBBPrivate GibberBytestreamIBBPrivate;
 struct _GibberBytestreamIBBPrivate
 {
-  GibberXmppConnection *xmpp_connnection;
-  GibberMucConnection *muc_connection;
+  GibberXmppConnection *xmpp_connection;
   gchar *self_id;
   gchar *peer_id;
   gchar *stream_id;
   gchar *stream_init_id;
-  GibberBytestreamIBBState state;
+  GibberBytestreamState state;
 
   guint16 seq;
   guint16 last_seq_recv;
+
+  gboolean dispose_has_run;
 };
 
 #define GIBBER_BYTESTREAM_IBB_GET_PRIVATE(obj) \
@@ -86,36 +95,14 @@ gibber_bytestream_ibb_init (GibberBytestreamIBB *self)
       GIBBER_TYPE_BYTESTREAM_IBB, GibberBytestreamIBBPrivate);
 
   self->priv = priv;
-
-  priv->xmpp_connnection = NULL;
-  priv->muc_connection = NULL;
-  priv->self_id = NULL;
-  priv->peer_id = NULL;
-
-  priv->seq = 0;
-  priv->last_seq_recv = 0;
 }
 
-static gboolean
-send_stanza (GibberBytestreamIBB *self,
-             GibberXmppStanza *stanza,
-             GError **error)
-{
-  GibberBytestreamIBBPrivate *priv = GIBBER_BYTESTREAM_IBB_GET_PRIVATE (self);
-
-  if (priv->muc_connection != NULL)
-    return gibber_muc_connection_send (priv->muc_connection, stanza, error);
-
-  if (priv->xmpp_connnection != NULL)
-    return gibber_xmpp_connection_send (priv->xmpp_connnection, stanza, error);
-
-  g_assert_not_reached ();
-}
-
-static gboolean
-stanza_received (GibberBytestreamIBB *self,
-                 GibberXmppStanza *stanza)
+static void
+xmpp_connection_received_stanza_cb (GibberXmppConnection *conn,
+                                    GibberXmppStanza *stanza,
+                                    gpointer user_data)
 {
+  GibberBytestreamIBB *self = (GibberBytestreamIBB *) user_data;
   GibberBytestreamIBBPrivate *priv = GIBBER_BYTESTREAM_IBB_GET_PRIVATE (self);
   GibberXmppNode *data;
   GString *str;
@@ -127,28 +114,28 @@ stanza_received (GibberBytestreamIBB *self,
       GIBBER_XMPP_NS_IBB);
   if (data == NULL)
     {
-      return FALSE;
+      return;
     }
 
   stream_id = gibber_xmpp_node_get_attribute (data, "sid");
   if (stream_id == NULL || strcmp (stream_id, priv->stream_id) != 0)
     {
       DEBUG ("bad stream id");
-      return FALSE;
+      return;
     }
 
-  if (priv->state != GIBBER_BYTESTREAM_IBB_STATE_OPEN)
+  if (priv->state != GIBBER_BYTESTREAM_STATE_OPEN)
     {
       DEBUG ("can't receive data through a not open bytestream (state: %d)",
           priv->state);
-      return FALSE;
+      return;
     }
 
   from = gibber_xmpp_node_get_attribute (stanza->node, "from");
   if (from == NULL)
     {
       DEBUG ("got a message without a from field, ignoring");
-      return FALSE;
+      return;
     }
 
   // XXX check sequence number ?
@@ -159,28 +146,6 @@ stanza_received (GibberBytestreamIBB *self,
 
   g_string_free (str, TRUE);
   g_free (decoded);
-  return TRUE;
-}
-
-static void
-xmpp_connection_received_stanza_cb (GibberXmppConnection *conn,
-                                    GibberXmppStanza *stanza,
-                                    gpointer user_data)
-{
-  GibberBytestreamIBB *self = (GibberBytestreamIBB *) user_data;
-
-  stanza_received (self, stanza);
-}
-
-static void
-muc_connection_received_stanza_cb (GibberMucConnection *conn,
-                                   const gchar *from,
-                                   GibberXmppStanza *stanza,
-                                   gpointer user_data)
-{
-  GibberBytestreamIBB *self = (GibberBytestreamIBB *) user_data;
-
-  stanza_received (self, stanza);
 }
 
 static void
@@ -189,9 +154,13 @@ gibber_bytestream_ibb_dispose (GObject *object)
   GibberBytestreamIBB *self = GIBBER_BYTESTREAM_IBB (object);
   GibberBytestreamIBBPrivate *priv = GIBBER_BYTESTREAM_IBB_GET_PRIVATE (self);
 
-  if (priv->state != GIBBER_BYTESTREAM_IBB_STATE_CLOSED)
+  if (priv->dispose_has_run)
+    return;
+  priv->dispose_has_run = TRUE;
+
+  if (priv->state != GIBBER_BYTESTREAM_STATE_CLOSED)
     {
-      gibber_bytestream_ibb_close (self);
+      gibber_bytestream_iface_close (GIBBER_BYTESTREAM_IFACE (self), NULL);
     }
 
   G_OBJECT_CLASS (gibber_bytestream_ibb_parent_class)->dispose (object);
@@ -204,10 +173,7 @@ gibber_bytestream_ibb_finalize (GObject *object)
   GibberBytestreamIBBPrivate *priv = GIBBER_BYTESTREAM_IBB_GET_PRIVATE (self);
 
   g_free (priv->stream_id);
-
-  if (priv->stream_init_id)
-    g_free (priv->stream_init_id);
-
+  g_free (priv->stream_init_id);
   g_free (priv->self_id);
   g_free (priv->peer_id);
 
@@ -226,10 +192,7 @@ gibber_bytestream_ibb_get_property (GObject *object,
   switch (property_id)
     {
       case PROP_XMPP_CONNECTION:
-        g_value_set_object (value, priv->xmpp_connnection);
-        break;
-      case PROP_MUC_CONNECTION:
-        g_value_set_object (value, priv->muc_connection);
+        g_value_set_object (value, priv->xmpp_connection);
         break;
       case PROP_SELF_ID:
         g_value_set_string (value, priv->self_id);
@@ -246,6 +209,9 @@ gibber_bytestream_ibb_get_property (GObject *object,
       case PROP_STATE:
         g_value_set_uint (value, priv->state);
         break;
+      case PROP_PROTOCOL:
+        g_value_set_string (value, GIBBER_XMPP_NS_IBB);
+        break;
       default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
         break;
@@ -264,17 +230,11 @@ gibber_bytestream_ibb_set_property (GObject *object,
   switch (property_id)
     {
       case PROP_XMPP_CONNECTION:
-        priv->xmpp_connnection = g_value_get_object (value);
-        if (priv->xmpp_connnection != NULL)
-          g_signal_connect (priv->xmpp_connnection, "received-stanza",
+        priv->xmpp_connection = g_value_get_object (value);
+        if (priv->xmpp_connection != NULL)
+          g_signal_connect (priv->xmpp_connection, "received-stanza",
               G_CALLBACK (xmpp_connection_received_stanza_cb), self);
         break;
-      case PROP_MUC_CONNECTION:
-        priv->muc_connection = g_value_get_object (value);
-        if (priv->muc_connection != NULL)
-          g_signal_connect (priv->muc_connection, "received-stanza",
-              G_CALLBACK (muc_connection_received_stanza_cb), self);
-        break;
       case PROP_SELF_ID:
         g_free (priv->self_id);
         priv->self_id = g_value_dup_string (value);
@@ -317,8 +277,8 @@ gibber_bytestream_ibb_constructor (GType type,
 
   priv = GIBBER_BYTESTREAM_IBB_GET_PRIVATE (GIBBER_BYTESTREAM_IBB (obj));
 
-  /* We can't be a private *and* a muc bytestream */
-  g_assert (priv->xmpp_connnection == NULL || priv->muc_connection == NULL);
+  g_assert (priv->xmpp_connection != NULL);
+  g_assert (priv->stream_init_id != NULL);
   g_assert (priv->self_id != NULL);
   g_assert (priv->peer_id != NULL);
 
@@ -342,6 +302,17 @@ gibber_bytestream_ibb_class_init (
   object_class->set_property = gibber_bytestream_ibb_set_property;
   object_class->constructor = gibber_bytestream_ibb_constructor;
 
+  g_object_class_override_property (object_class, PROP_SELF_ID,
+      "self-id");
+  g_object_class_override_property (object_class, PROP_PEER_ID,
+      "peer-id");
+  g_object_class_override_property (object_class, PROP_STREAM_ID,
+      "stream-id");
+  g_object_class_override_property (object_class, PROP_STATE,
+      "state");
+  g_object_class_override_property (object_class, PROP_PROTOCOL,
+      "protocol");
+
   param_spec = g_param_spec_object (
       "xmpp-connection",
       "GibberXmppConnection object",
@@ -356,56 +327,6 @@ gibber_bytestream_ibb_class_init (
   g_object_class_install_property (object_class, PROP_XMPP_CONNECTION,
       param_spec);
 
-  param_spec = g_param_spec_object (
-      "muc-connection",
-      "GibberMucConnection object",
-      "Gibber MUC connection object used for communication by this "
-      "bytestream if it's a muc one",
-      GIBBER_TYPE_MUC_CONNECTION,
-      G_PARAM_CONSTRUCT_ONLY |
-      G_PARAM_READWRITE |
-      G_PARAM_STATIC_NAME |
-      G_PARAM_STATIC_NICK |
-      G_PARAM_STATIC_BLURB);
-  g_object_class_install_property (object_class, PROP_MUC_CONNECTION,
-      param_spec);
-
- param_spec = g_param_spec_string (
-      "self-id",
-      "self ID",
-      "the ID of the local user",
-      "",
-      G_PARAM_CONSTRUCT_ONLY |
-      G_PARAM_READWRITE |
-      G_PARAM_STATIC_NAME |
-      G_PARAM_STATIC_NICK |
-      G_PARAM_STATIC_BLURB);
-  g_object_class_install_property (object_class, PROP_SELF_ID, param_spec);
-
- param_spec = g_param_spec_string (
-      "peer-id",
-      "peer JID",
-      "the ID of the muc or the remote user associated with this bytestream",
-      "",
-      G_PARAM_CONSTRUCT_ONLY |
-      G_PARAM_READWRITE |
-      G_PARAM_STATIC_NAME |
-      G_PARAM_STATIC_NICK |
-      G_PARAM_STATIC_BLURB);
-  g_object_class_install_property (object_class, PROP_PEER_ID, param_spec);
-
-  param_spec = g_param_spec_string (
-      "stream-id",
-      "stream ID",
-      "the ID of the stream",
-      "",
-      G_PARAM_CONSTRUCT_ONLY |
-      G_PARAM_READWRITE |
-      G_PARAM_STATIC_NAME |
-      G_PARAM_STATIC_NICK |
-      G_PARAM_STATIC_BLURB);
-  g_object_class_install_property (object_class, PROP_STREAM_ID, param_spec);
-
   param_spec = g_param_spec_string (
       "stream-init-id",
       "stream init ID",
@@ -419,19 +340,6 @@ gibber_bytestream_ibb_class_init (
   g_object_class_install_property (object_class, PROP_STREAM_INIT_ID,
       param_spec);
 
-  param_spec = g_param_spec_uint (
-      "state",
-      "Bytestream state",
-      "An enum (BytestreamIBBState) signifying the current state of"
-      "this bytestream object",
-      0, LAST_GIBBER_BYTESTREAM_IBB_STATE - 1,
-      GIBBER_BYTESTREAM_IBB_STATE_LOCAL_PENDING,
-      G_PARAM_READWRITE |
-      G_PARAM_STATIC_NAME |
-      G_PARAM_STATIC_NICK |
-      G_PARAM_STATIC_BLURB);
-  g_object_class_install_property (object_class, PROP_STATE, param_spec);
-
   signals[DATA_RECEIVED] =
     g_signal_new ("data-received",
                   G_OBJECT_CLASS_TYPE (gibber_bytestream_ibb_class),
@@ -451,19 +359,23 @@ gibber_bytestream_ibb_class_init (
                   G_TYPE_NONE, 1, G_TYPE_UINT);
 }
 
-gboolean
-send_data_to (GibberBytestreamIBB *self,
-              const gchar *to,
-              gboolean groupchat,
-              guint len,
-              gchar *str)
+/*
+ * gibber_bytestream_ibb_send
+ *
+ * Implements gibber_bytestream_iface_send on GibberBytestreamIface
+ */
+static gboolean
+gibber_bytestream_ibb_send (GibberBytestreamIface *bytestream,
+                            guint len,
+                            const gchar *str)
 {
+  GibberBytestreamIBB *self = GIBBER_BYTESTREAM_IBB (bytestream);
   GibberBytestreamIBBPrivate *priv = GIBBER_BYTESTREAM_IBB_GET_PRIVATE (self);
   GibberXmppStanza *stanza;
   gchar *seq, *encoded;
   gboolean ret;
 
-  if (priv->state != GIBBER_BYTESTREAM_IBB_STATE_OPEN)
+  if (priv->state != GIBBER_BYTESTREAM_STATE_OPEN)
     {
       DEBUG ("can't send data through a not open bytestream (state: %d)",
           priv->state);
@@ -498,13 +410,8 @@ send_data_to (GibberBytestreamIBB *self,
       GIBBER_NODE_END,
       GIBBER_STANZA_END);
 
-  if (groupchat)
-    {
-      gibber_xmpp_node_set_attribute (stanza->node, "type", "groupchat");
-    }
-
   DEBUG ("send %d bytes", len);
-  ret = send_stanza (self, stanza, NULL);
+  ret = gibber_xmpp_connection_send (priv->xmpp_connection, stanza, NULL);
 
   g_object_unref (stanza);
   g_free (encoded);
@@ -513,126 +420,125 @@ send_data_to (GibberBytestreamIBB *self,
   return ret;
 }
 
-gboolean
-gibber_bytestream_ibb_send (GibberBytestreamIBB *self,
-                            guint len,
-                            gchar *str)
+static GibberXmppStanza *
+create_si_accept_iq (GibberBytestreamIBB *self)
 {
-  GibberBytestreamIBBPrivate *priv;
-  gboolean groupchat = FALSE;
+  GibberBytestreamIBBPrivate *priv = GIBBER_BYTESTREAM_IBB_GET_PRIVATE (self);
 
-  g_assert (GIBBER_IS_BYTESTREAM_IBB (self));
-  priv = GIBBER_BYTESTREAM_IBB_GET_PRIVATE (self);
-
-  if (priv->muc_connection != NULL)
-    groupchat = TRUE;
-
-  return send_data_to (self, priv->peer_id, groupchat, len, str);
+  return gibber_xmpp_stanza_build (
+      GIBBER_STANZA_TYPE_IQ, GIBBER_STANZA_SUB_TYPE_RESULT,
+      priv->self_id, priv->peer_id,
+      GIBBER_NODE_ATTRIBUTE, "id", priv->stream_init_id,
+      GIBBER_NODE, "si",
+        GIBBER_NODE_XMLNS, GIBBER_XMPP_NS_SI,
+        GIBBER_NODE, "feature",
+          GIBBER_NODE_XMLNS, GIBBER_XMPP_NS_FEATURENEG,
+          GIBBER_NODE, "x",
+            GIBBER_NODE_XMLNS, GIBBER_XMPP_NS_DATA,
+            GIBBER_NODE_ATTRIBUTE, "type", "submit",
+            GIBBER_NODE, "field",
+              GIBBER_NODE_ATTRIBUTE, "var", "stream-method",
+              GIBBER_NODE, "value",
+                GIBBER_NODE_TEXT, GIBBER_XMPP_NS_IBB,
+              GIBBER_NODE_END,
+            GIBBER_NODE_END,
+          GIBBER_NODE_END,
+        GIBBER_NODE_END,
+      GIBBER_NODE_END, GIBBER_STANZA_END);
 }
 
 /*
-GibberXmppStanza *
-gibber_bytestream_ibb_make_accept_iq (GibberBytestreamIBB *self)
+ * gibber_bytestream_ibb_accept
+ *
+ * Implements gibber_bytestream_iface_accept on GibberBytestreamIface
+ */
+static void
+gibber_bytestream_ibb_accept (GibberBytestreamIface *bytestream,
+                              GibberBytestreamAugmentSiAcceptReply func,
+                              gpointer user_data)
 {
+  GibberBytestreamIBB *self = GIBBER_BYTESTREAM_IBB (bytestream);
   GibberBytestreamIBBPrivate *priv = GIBBER_BYTESTREAM_IBB_GET_PRIVATE (self);
-  LmMessage *msg;
+  GibberXmppStanza *stanza;
+  GibberXmppNode *si;
 
-  if (priv->peer_handle_type == TP_HANDLE_TYPE_ROOM ||
-      priv->stream_init_id == NULL)
-    {
-      DEBUG ("bytestream was not created due to a SI request");
-      return NULL;
-    }
-
-  msg = gibber_bytestream_factory_make_accept_iq (priv->peer_jid,
-      priv->stream_init_id, NS_IBB);
-
-  return msg;
-}
-*/
-
-void
-gibber_bytestream_ibb_accept (GibberBytestreamIBB *self, GibberXmppStanza *msg)
-{
-  GibberBytestreamIBBPrivate *priv = GIBBER_BYTESTREAM_IBB_GET_PRIVATE (self);
-  GError *error = NULL;
-
-  if (priv->state != GIBBER_BYTESTREAM_IBB_STATE_LOCAL_PENDING)
+  if (priv->state != GIBBER_BYTESTREAM_STATE_LOCAL_PENDING)
     {
       /* The stream was previoulsy or automatically accepted */
+      DEBUG ("stream was already accepted");
       return;
     }
 
-  if (priv->xmpp_connnection != NULL ||
-      priv->stream_init_id == NULL)
+  stanza = create_si_accept_iq (self);
+  si = gibber_xmpp_node_get_child_ns (stanza->node, "si", GIBBER_XMPP_NS_SI);
+  g_assert (si != NULL);
+
+  if (func != NULL)
     {
-      DEBUG ("can't accept a bytestream not created due to a SI request");
-      return;
+      /* let the caller add his profile specific data */
+      func (si, user_data);
     }
 
-  if (send_stanza (self, msg, &error))
-    {
-      priv->state = GIBBER_BYTESTREAM_IBB_STATE_ACCEPTED;
-    }
-  else
-    {
-      DEBUG ("send accept stanza failed: %s", error->message);
-      g_error_free (error);
-    }
+  gibber_xmpp_connection_send (priv->xmpp_connection, stanza, NULL);
+
+  DEBUG ("stream is now accepted");
+  g_object_set (self, "state", GIBBER_BYTESTREAM_STATE_ACCEPTED, NULL);
 }
 
 static void
-gibber_bytestream_ibb_decline (GibberBytestreamIBB *self)
+gibber_bytestream_ibb_decline (GibberBytestreamIBB *self,
+                               GError *error)
 {
-/*
   GibberBytestreamIBBPrivate *priv = GIBBER_BYTESTREAM_IBB_GET_PRIVATE (self);
-  GibberXmppStanza *stanza
+  GibberXmppStanza *stanza;
 
-  if (priv->state != GIBBER_BYTESTREAM_IBB_STATE_LOCAL_PENDING)
+  g_return_if_fail (priv->state == GIBBER_BYTESTREAM_STATE_LOCAL_PENDING);
+
+  stanza = gibber_xmpp_stanza_build (
+      GIBBER_STANZA_TYPE_IQ, GIBBER_STANZA_SUB_TYPE_ERROR,
+      priv->self_id, priv->peer_id,
+      GIBBER_NODE_ATTRIBUTE, "id", priv->stream_init_id,
+      GIBBER_STANZA_END);
+
+  if (error != NULL && error->domain == GIBBER_XMPP_ERROR)
     {
-      DEBUG ("bytestream is not in the local pending state (state %d)",
-          priv->state);
-      return;
+      gibber_xmpp_error_to_node (error->code, stanza->node, error->message);
+    }
+  else
+    {
+      gibber_xmpp_error_to_node (XMPP_ERROR_FORBIDDEN, stanza->node,
+          "Offer Declined");
     }
 
-  if (priv->peer_handle_type == TP_HANDLE_TYPE_ROOM ||
-      priv->stream_init_id == NULL)
-    {
-      DEBUG ("can't decline a bytestream not created due to a SI request");
-      return;
-    }
+  gibber_xmpp_connection_send (priv->xmpp_connection, stanza, NULL);
 
-  msg = gibber_bytestream_factory_make_decline_iq (priv->peer_jid,
-      priv->stream_init_id);
-
-  _gibber_connection_send (priv->conn, msg, NULL);
-
-  lm_message_unref (msg);
-  */
+  g_object_unref (stanza);
 }
 
-void
-gibber_bytestream_ibb_close (GibberBytestreamIBB *self)
+/*
+ * gibber_bytestream_ibb_close
+ *
+ * Implements gibber_bytestream_iface_close on GibberBytestreamIface
+ */
+static void
+gibber_bytestream_ibb_close (GibberBytestreamIface *bytestream,
+                             GError *error)
 {
+  GibberBytestreamIBB *self = GIBBER_BYTESTREAM_IBB (bytestream);
   GibberBytestreamIBBPrivate *priv = GIBBER_BYTESTREAM_IBB_GET_PRIVATE (self);
 
-  if (priv->state == GIBBER_BYTESTREAM_IBB_STATE_CLOSED)
+  if (priv->state == GIBBER_BYTESTREAM_STATE_CLOSED)
      /* bytestream already closed, do nothing */
      return;
 
-  if (priv->state == GIBBER_BYTESTREAM_IBB_STATE_LOCAL_PENDING)
+  if (priv->state == GIBBER_BYTESTREAM_STATE_LOCAL_PENDING)
     {
-      if (priv->stream_init_id != NULL)
-        {
-          /* Stream was created using SI so we decline the request */
-          gibber_bytestream_ibb_decline (self);
-        }
+      /* Stream was created using SI so we decline the request */
+      gibber_bytestream_ibb_decline (self, error);
     }
 
-  else if (priv->xmpp_connnection != NULL)
+  else if (priv->xmpp_connection != NULL)
     {
-      /* XXX : Does it make sense to send a close message in a
-       * muc bytestream ? */
       GibberXmppStanza *stanza;
 
       DEBUG ("send IBB close stanza");
@@ -645,12 +551,12 @@ gibber_bytestream_ibb_close (GibberBytestreamIBB *self)
             GIBBER_NODE_ATTRIBUTE, "sid", priv->stream_id,
           GIBBER_NODE_END, GIBBER_STANZA_END);
 
-      send_stanza (self, stanza, NULL);
+      gibber_xmpp_connection_send (priv->xmpp_connection, stanza, NULL);
 
       g_object_unref (stanza);
     }
 
-  g_object_set (self, "state", GIBBER_BYTESTREAM_IBB_STATE_CLOSED, NULL);
+  g_object_set (self, "state", GIBBER_BYTESTREAM_STATE_CLOSED, NULL);
 }
 
 #if 0
@@ -667,33 +573,39 @@ ibb_init_reply_cb (GibberConnection *conn,
     {
       /* yeah, stream initiated */
       DEBUG ("IBB stream initiated");
-      g_object_set (self, "state", GIBBER_BYTESTREAM_IBB_STATE_OPEN, NULL);
+      g_object_set (self, "state", GIBBER_BYTESTREAM_STATE_OPEN, NULL);
     }
   else
     {
       DEBUG ("error during IBB initiation");
-      g_object_set (self, "state", GIBBER_BYTESTREAM_IBB_STATE_CLOSED, NULL);
+      g_object_set (self, "state", GIBBER_BYTESTREAM_STATE_CLOSED, NULL);
     }
 
   return LM_HANDLER_RESULT_REMOVE_MESSAGE;
 }
 #endif
 
-gboolean
-gibber_bytestream_ibb_initiation (GibberBytestreamIBB *self)
+/*
+ * gibber_bytestream_ibb_initiate
+ *
+ * Implements gibber_bytestream_iface_initiate on GibberBytestreamIface
+ */
+static gboolean
+gibber_bytestream_ibb_initiate (GibberBytestreamIface *bytestream)
 {
+  GibberBytestreamIBB *self = GIBBER_BYTESTREAM_IBB (bytestream);
   GibberBytestreamIBBPrivate *priv = GIBBER_BYTESTREAM_IBB_GET_PRIVATE (self);
   GibberXmppStanza *msg;
   GError *error = NULL;
 
-  if (priv->state != GIBBER_BYTESTREAM_IBB_STATE_INITIATING)
+  if (priv->state != GIBBER_BYTESTREAM_STATE_INITIATING)
     {
       DEBUG ("bytestream is not is the initiating state (state %d",
           priv->state);
       return FALSE;
     }
 
-  if (priv->xmpp_connnection == NULL)
+  if (priv->xmpp_connection == NULL)
     {
       DEBUG ("Can only initiate a private bytestream");
       return FALSE;
@@ -715,7 +627,7 @@ gibber_bytestream_ibb_initiation (GibberBytestreamIBB *self)
       GIBBER_NODE_END, GIBBER_STANZA_END);
 
   /* XXX should send using _with_reply (ibb_init_reply_cb) */
-  if (!send_stanza (self, msg, &error))
+  if (!gibber_xmpp_connection_send (priv->xmpp_connection, msg, &error))
     {
       DEBUG ("Error when sending IBB init stanza: %s", error->message);
 
@@ -728,3 +640,15 @@ gibber_bytestream_ibb_initiation (GibberBytestreamIBB *self)
 
   return TRUE;
 }
+
+static void
+bytestream_iface_init (gpointer g_iface,
+                       gpointer iface_data)
+{
+  GibberBytestreamIfaceClass *klass = (GibberBytestreamIfaceClass *) g_iface;
+
+  klass->initiate = gibber_bytestream_ibb_initiate;
+  klass->send = gibber_bytestream_ibb_send;
+  klass->close = gibber_bytestream_ibb_close;
+  klass->accept = gibber_bytestream_ibb_accept;
+}
