@@ -21,6 +21,8 @@
 #ifndef __SALUT_CONNECTION_H__
 #define __SALUT_CONNECTION_H__
 
+#include "config.h"
+
 #include <glib-object.h>
 #include <dbus/dbus-glib.h>
 
@@ -29,6 +31,8 @@
 #include <telepathy-glib/presence-mixin.h>
 #include <telepathy-glib/svc-connection.h>
 
+#include <gibber/gibber-xmpp-stanza.h>
+
 G_BEGIN_DECLS
 
 typedef struct _SalutConnection SalutConnection;
@@ -74,6 +78,16 @@ typedef enum {
   LIST_HANDLE_LAST = LIST_HANDLE_KNOWN
 } SalutConnectionListHandle;
 
+#ifdef ENABLE_OLPC
+void
+salut_connection_olpc_observe_invitation (SalutConnection *connection,
+    TpHandle room, TpHandle invitor_handle, GibberXmppNode *invite_node);
+
+gboolean
+salut_connection_olpc_observe_muc_stanza (SalutConnection *self, TpHandle room,
+    TpHandle sender, GibberXmppStanza *stanza);
+#endif
+
 G_END_DECLS
 
 #endif /* #ifndef __SALUT_CONNECTION_H__*/
