@@ -30,8 +30,8 @@
 #include "salut-contact.h"
 #include "salut-presence-enumtypes.h"
 
-#include "salut-avahi-service-browser.h"
-#include "salut-avahi-enums.h"
+#include <avahi-gobject/ga-service-browser.h>
+#include <avahi-gobject/ga-enums.h>
 
 #include <telepathy-glib/channel-factory-iface.h>
 #include <telepathy-glib/interfaces.h>
@@ -57,6 +57,9 @@ G_DEFINE_TYPE_WITH_CODE(SalutContactManager, salut_contact_manager,
 enum
 {
   CONTACT_CHANGE,
+#ifdef ENABLE_OLPC
+  ACTIVITY_PROPERTIES_CHANGE,
+#endif
   LAST_SIGNAL
 };
 
@@ -70,6 +73,8 @@ typedef struct
   gchar *color;
   gchar *name;
   gchar *type;
+  gchar *tags;
+  gboolean is_private;
   size_t refcount;
 } SalutContactManagerActivity;
 
@@ -78,14 +83,14 @@ typedef struct _SalutContactManagerPrivate SalutContactManagerPrivate;
 
 struct _SalutContactManagerPrivate
 {
-  SalutAvahiServiceBrowser *presence_browser;
+  GaServiceBrowser *presence_browser;
 #ifdef ENABLE_OLPC
-  SalutAvahiServiceBrowser *activity_browser;
+  GaServiceBrowser *activity_browser;
   GHashTable *olpc_activities_by_mdns;
   GHashTable *olpc_activities_by_room;
 #endif
   SalutConnection *connection;
-  SalutAvahiClient *client;
+  GaClient *client;
   GHashTable *channels;
   GHashTable *contacts;
   gboolean dispose_has_run;
@@ -108,6 +113,7 @@ activity_new (SalutContactManager *mgr,
   tp_handle_ref (room_repo, room_handle);
   activity->room = room_handle;
   activity->mgr = mgr;
+  activity->is_private = TRUE;
 
   DEBUG ("Creating SalutContactManagerActivity: handle %u", room_handle);
   g_hash_table_insert (priv->olpc_activities_by_room,
@@ -138,6 +144,7 @@ activity_unref (SalutContactManagerActivity *activity)
   g_free (activity->color);
   g_free (activity->name);
   g_free (activity->type);
+  g_free (activity->tags);
   g_slice_free (SalutContactManagerActivity, activity);
 }
 #endif
@@ -183,6 +190,19 @@ salut_contact_manager_class_init (SalutContactManagerClass *salut_contact_manage
       G_TYPE_NONE, 2,
       SALUT_TYPE_CONTACT,
       G_TYPE_INT);
+
+#ifdef ENABLE_OLPC
+  signals[ACTIVITY_PROPERTIES_CHANGE] = g_signal_new(
+      "activity-properties-change",
+      G_OBJECT_CLASS_TYPE(salut_contact_manager_class),
+      G_SIGNAL_RUN_LAST,
+      0,
+      NULL, NULL,
+      salut_signals_marshal_VOID__UINT_STRING_STRING_STRING_STRING_STRING_BOOLEAN,
+      G_TYPE_NONE, 7,
+      G_TYPE_UINT, G_TYPE_STRING, G_TYPE_STRING,
+      G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_BOOLEAN);
+#endif
 }
 
 static gboolean
@@ -248,75 +268,24 @@ change_all_groups(SalutContactManager *mgr, TpIntSet *add, TpIntSet *rem) {
 static void
 contact_found_cb(SalutContact *contact, gpointer userdata) {
   SalutContactManager *mgr = SALUT_CONTACT_MANAGER(userdata);
-  SalutContactManagerPrivate *priv = SALUT_CONTACT_MANAGER_GET_PRIVATE(mgr);
-  TpHandleRepoIface *handle_repo = tp_base_connection_get_handles(
-      TP_BASE_CONNECTION(priv->connection), TP_HANDLE_TYPE_CONTACT);
-
   TpIntSet *to_add = tp_intset_new();
   TpIntSet *to_rem = tp_intset_new();
-  TpHandle handle;
 
-  handle = tp_handle_ensure(handle_repo, contact->name, NULL, NULL);
-  tp_intset_add(to_add, handle);
+  tp_intset_add(to_add, contact->handle);
   change_all_groups(mgr, to_add, to_rem);
-  /* Add an extra ref, to ensure keeping this untill we got the lost signal */
   tp_intset_destroy(to_add);
   tp_intset_destroy(to_rem);
 }
 
 #ifdef ENABLE_OLPC
-static void
-activity_change_cb(SalutContact *contact,
-                   const gchar *service_name,
-                   TpHandle room_handle,
-                   const gchar *activity_id,
-                   const gchar *color,
-                   const gchar *name,
-                   const gchar *type,
-                   gpointer userdata)
+static gboolean
+update_activity (SalutContactManagerActivity *activity,
+                 const gchar *name,
+                 const gchar *type,
+                 const gchar *color,
+                 const gchar *tags)
 {
-  SalutContactManager *mgr = SALUT_CONTACT_MANAGER (userdata);
-  SalutContactManagerPrivate *priv = SALUT_CONTACT_MANAGER_GET_PRIVATE (mgr);
   gboolean changed = FALSE;
-  SalutContactManagerActivity *activity;
-
-  DEBUG ("enter: sn=%s, h=%u, aid=%s, c=%s, n=%s, t=%s",
-      service_name, room_handle,
-      activity_id ? activity_id : "<NULL>",
-      color ? color : "<NULL>",
-      name ? name : "<NULL>",
-      type ? type : "<NULL>");
-
-  if (room_handle == 0)
-    {
-      DEBUG ("Treating activity advertisement with room_handle 0 as removal");
-      g_hash_table_remove (priv->olpc_activities_by_mdns, name);
-      return;
-    }
-
-  activity = g_hash_table_lookup (priv->olpc_activities_by_mdns, service_name);
-  if (activity == NULL)
-    {
-      activity = g_hash_table_lookup (priv->olpc_activities_by_room,
-          GUINT_TO_POINTER (room_handle));
-      if (activity == NULL)
-        {
-          activity = activity_new (mgr, room_handle);
-          changed = TRUE;
-        }
-      else
-        {
-          activity->refcount++;
-        }
-      DEBUG ("Activity %u now advertised as %s", room_handle, service_name);
-      g_hash_table_insert (priv->olpc_activities_by_mdns,
-          g_strdup (service_name), activity);
-    }
-  else
-    {
-      DEBUG ("Activity %u already known to be advertised as %s", room_handle,
-          service_name);
-    }
 
   if (name != NULL && tp_strdiff (activity->name, name))
     {
@@ -338,6 +307,94 @@ activity_change_cb(SalutContact *contact,
       activity->color = g_strdup (color);
       changed = TRUE;
     }
+
+  if (tags != NULL && tp_strdiff (activity->tags, tags))
+    {
+      g_free (activity->tags);
+      activity->tags = g_strdup (tags);
+      changed = TRUE;
+    }
+
+  return changed;
+}
+
+static void
+activity_change_cb(SalutContact *contact,
+                   const gchar *service_name,
+                   TpHandle room_handle,
+                   const gchar *activity_id,
+                   const gchar *color,
+                   const gchar *name,
+                   const gchar *type,
+                   const gchar *tags,
+                   gpointer userdata)
+{
+  SalutContactManager *mgr = SALUT_CONTACT_MANAGER (userdata);
+  SalutContactManagerPrivate *priv = SALUT_CONTACT_MANAGER_GET_PRIVATE (mgr);
+  gboolean changed = FALSE;
+  SalutContactManagerActivity *activity;
+
+  DEBUG ("enter: sn=%s, h=%u, aid=%s, c=%s, n=%s, t=%s, tags=%s",
+      service_name, room_handle,
+      activity_id ? activity_id : "<NULL>",
+      color ? color : "<NULL>",
+      name ? name : "<NULL>",
+      type ? type : "<NULL>",
+      tags ? tags : "<NULL>");
+
+  if (room_handle == 0)
+    {
+      DEBUG ("Treating activity advertisement with room_handle 0 as removal");
+      g_hash_table_remove (priv->olpc_activities_by_mdns, name);
+      return;
+    }
+
+  activity = g_hash_table_lookup (priv->olpc_activities_by_mdns, service_name);
+  if (activity == NULL)
+    {
+      activity = g_hash_table_lookup (priv->olpc_activities_by_room,
+          GUINT_TO_POINTER (room_handle));
+      if (activity == NULL)
+        {
+          activity = activity_new (mgr, room_handle);
+          /* we discover this activity, so it's not a private one */
+          activity->is_private = FALSE;
+          activity->activity_id = g_strdup (activity_id);
+          changed = TRUE;
+        }
+      else
+        {
+          if (activity->is_private)
+            {
+              /* We knew this activity before as a private activity.
+               * We don't increment it's refcount now as it was set to 1
+               * due to the invitation */
+              activity->is_private = FALSE;
+              changed = TRUE;
+            }
+          else
+            {
+              activity->refcount++;
+            }
+        }
+      DEBUG ("Activity %u now advertised as %s", room_handle, service_name);
+      g_hash_table_insert (priv->olpc_activities_by_mdns,
+          g_strdup (service_name), activity);
+    }
+  else
+    {
+      DEBUG ("Activity %u already known to be advertised as %s", room_handle,
+          service_name);
+    }
+
+  if (update_activity (activity, name, type, color, tags))
+    changed = TRUE;
+
+  if (changed)
+    {
+      g_signal_emit (mgr, signals[ACTIVITY_PROPERTIES_CHANGE], 0, room_handle,
+          activity_id, color, name, type, tags, FALSE);
+    }
 }
 #endif
 
@@ -353,20 +410,13 @@ contact_change_cb(SalutContact *contact, gint changes, gpointer userdata) {
 static void
 contact_lost_cb(SalutContact *contact, gpointer userdata) {
   SalutContactManager *mgr = SALUT_CONTACT_MANAGER(userdata);
-  SalutContactManagerPrivate *priv = SALUT_CONTACT_MANAGER_GET_PRIVATE(mgr);
-  TpHandleRepoIface *handle_repo = tp_base_connection_get_handles(
-      TP_BASE_CONNECTION(priv->connection), TP_HANDLE_TYPE_CONTACT);
 
   TpIntSet *to_add = tp_intset_new();
   TpIntSet *to_rem = tp_intset_new();
-  TpHandle handle;
 
   DEBUG("Removing %s from contacts", contact->name);
-  handle = tp_handle_lookup(handle_repo, contact->name, NULL, NULL);
 
-  g_assert(handle != 0);
-
-  tp_intset_add(to_rem, handle);
+  tp_intset_add(to_rem, contact->handle);
   change_all_groups(mgr, to_add, to_rem);
 
   tp_intset_destroy(to_add);
@@ -395,7 +445,9 @@ salut_contact_manager_merge_olpc_activity_properties
                                          TpHandle handle,
                                          const gchar **color,
                                          const gchar **name,
-                                         const gchar **type)
+                                         const gchar **type,
+                                         const gchar **tags,
+                                         gboolean *is_private)
 {
   SalutContactManagerPrivate *priv = SALUT_CONTACT_MANAGER_GET_PRIVATE (self);
   SalutContactManagerActivity *activity = g_hash_table_lookup (
@@ -410,6 +462,10 @@ salut_contact_manager_merge_olpc_activity_properties
     *name = activity->name;
   if (activity->type != NULL && type != NULL)
     *type = activity->type;
+  if (activity->tags != NULL && tags != NULL)
+    *tags = activity->tags;
+  if (is_private != NULL)
+    *is_private = activity->is_private;
   return TRUE;
 }
 
@@ -422,24 +478,113 @@ split_activity_name (const gchar **contact_name)
   if (*contact_name == NULL)
     {
       *contact_name = orig;
-      DEBUG ("Ignoring invalid _olpc-activity with no ':': %s", orig);
+      DEBUG ("Ignoring invalid OLPC activity DNS-SD with no ':': %s", orig);
       return FALSE;
     }
   (*contact_name)++;
   return TRUE;
 }
+
+void
+salut_contact_manager_add_invited_olpc_activity (SalutContactManager *self,
+                                                 SalutContact *inviter,
+                                                 TpHandle room,
+                                                 const gchar *activity_id,
+                                                 const gchar *color,
+                                                 const gchar *name,
+                                                 const gchar *type,
+                                                 const gchar *tags)
+{
+  SalutContactManagerPrivate *priv = SALUT_CONTACT_MANAGER_GET_PRIVATE (self);
+  SalutContactManagerActivity *activity;
+  gboolean changed = FALSE;
+
+  activity = g_hash_table_lookup (priv->olpc_activities_by_room,
+      GUINT_TO_POINTER (room));
+  if (activity == NULL)
+    {
+      DEBUG ("new activity %s created due to invite from %s", activity_id,
+          inviter->name);
+
+      activity = activity_new (self, room);
+      activity->activity_id = g_strdup (activity_id);
+      /* We didn't know this activity before, it's a private one */
+      activity->is_private = TRUE;
+      g_hash_table_insert (priv->olpc_activities_by_room,
+          GUINT_TO_POINTER (room), activity);
+      changed = TRUE;
+    }
+
+  if (update_activity (activity, name, type, color, tags))
+    changed = TRUE;
+
+  if (changed)
+    {
+      g_signal_emit (self, signals[ACTIVITY_PROPERTIES_CHANGE], 0, room,
+          activity_id, color, name, type, tags, TRUE);
+    }
+}
 #endif
 
+static SalutContact *
+salut_contact_manager_create_contact (SalutContactManager *self,
+                                      const gchar *name)
+{
+  SalutContactManagerPrivate *priv = SALUT_CONTACT_MANAGER_GET_PRIVATE (self);
+  SalutContact *contact;
+
+  g_assert (g_hash_table_lookup (priv->contacts, name) == NULL);
+
+  contact = salut_contact_new (priv->client, priv->connection, name);
+
+  g_hash_table_insert (priv->contacts, g_strdup (contact->name), contact);
+  DEBUG("Adding %s to contacts", name);
+
+  g_signal_connect (contact, "found",
+      G_CALLBACK(contact_found_cb), self);
+  g_signal_connect (contact, "contact-change",
+      G_CALLBACK(contact_change_cb), self);
+#ifdef ENABLE_OLPC
+  g_signal_connect (contact, "activity-change",
+      G_CALLBACK (activity_change_cb), self);
+#endif
+  g_signal_connect (contact, "lost",
+      G_CALLBACK(contact_lost_cb), self);
+
+  g_object_weak_ref (G_OBJECT (contact), _contact_finalized_cb , self);
+
+  return contact;
+}
+
+SalutContact *
+salut_contact_manager_ensure_contact (SalutContactManager *self,
+                                      const gchar *name)
+{
+  SalutContactManagerPrivate *priv = SALUT_CONTACT_MANAGER_GET_PRIVATE (self);
+  SalutContact *contact;
+
+  contact = g_hash_table_lookup (priv->contacts, name);
+  if (contact == NULL)
+    {
+      DEBUG ("contact %s doesn't exist yet. Creating it", name);
+      contact = salut_contact_manager_create_contact (self, name);
+    }
+  else
+    {
+      g_object_ref (contact);
+    }
+
+  return contact;
+}
+
 static void
-browser_found(SalutAvahiServiceBrowser *browser,
+browser_found(GaServiceBrowser *browser,
               AvahiIfIndex interface, AvahiProtocol protocol,
               const char *name, const char *type, const char *domain,
-              SalutAvahiLookupResultFlags flags,
+              GaLookupResultFlags flags,
               gpointer userdata) {
   SalutContactManager *mgr = SALUT_CONTACT_MANAGER(userdata);
   SalutContactManagerPrivate *priv = SALUT_CONTACT_MANAGER_GET_PRIVATE(mgr);
-  TpHandleRepoIface *room_repo = tp_base_connection_get_handles
-      ((TpBaseConnection *) priv->connection, TP_HANDLE_TYPE_ROOM);
   SalutContact *contact;
   const char *contact_name = name;
 
@@ -457,20 +602,7 @@ browser_found(SalutAvahiServiceBrowser *browser,
   /* FIXME: For now we assume name is unique on the lan */
   contact = g_hash_table_lookup (priv->contacts, contact_name);
   if (contact == NULL) {
-    contact = salut_contact_new(priv->client, room_repo, contact_name);
-    g_hash_table_insert(priv->contacts, g_strdup(contact->name), contact);
-    DEBUG("Adding %s to contacts", contact_name);
-    g_signal_connect(contact, "found",
-                     G_CALLBACK(contact_found_cb), mgr);
-    g_signal_connect(contact, "contact-change",
-                     G_CALLBACK(contact_change_cb), mgr);
-#ifdef ENABLE_OLPC
-    g_signal_connect(contact, "activity-change",
-        G_CALLBACK(activity_change_cb), mgr);
-#endif
-    g_signal_connect(contact, "lost",
-                     G_CALLBACK(contact_lost_cb), mgr);
-    g_object_weak_ref(G_OBJECT(contact), _contact_finalized_cb , mgr);
+    contact = salut_contact_manager_create_contact (mgr, contact_name);
   } else  if (!salut_contact_has_services(contact)) {
     g_object_ref(contact);
   }
@@ -478,10 +610,10 @@ browser_found(SalutAvahiServiceBrowser *browser,
 }
 
 static void
-browser_removed(SalutAvahiServiceBrowser *browser,
+browser_removed(GaServiceBrowser *browser,
               AvahiIfIndex interface, AvahiProtocol protocol,
               const char *name, const char *type, const char *domain,
-              SalutAvahiLookupResultFlags flags,
+              GaLookupResultFlags flags,
               gpointer userdata) {
   SalutContactManager *mgr = SALUT_CONTACT_MANAGER(userdata);
   SalutContactManagerPrivate *priv = SALUT_CONTACT_MANAGER_GET_PRIVATE(mgr);
@@ -508,12 +640,12 @@ browser_removed(SalutAvahiServiceBrowser *browser,
     salut_contact_remove_service(contact, interface, protocol,
                                  name, type, domain);
   } else {
-    g_message("Unknown contact removed from service browser");
+    DEBUG ("Unknown contact removed from service browser");
   }
 }
 
 static void
-browser_failed(SalutAvahiServiceBrowser *browser,
+browser_failed(GaServiceBrowser *browser,
                GError *error, gpointer userdata) {
   /* FIXME proper error handling */
   g_warning("browser failed -> %s", error->message);
@@ -724,10 +856,11 @@ salut_contact_manager_new(SalutConnection *connection) {
 
   priv->connection = connection;
 
-  priv->presence_browser = salut_avahi_service_browser_new ("_presence._tcp");
+  priv->presence_browser = ga_service_browser_new
+      (SALUT_DNSSD_PRESENCE);
 #ifdef ENABLE_OLPC
-  priv->activity_browser = salut_avahi_service_browser_new
-    ("_olpc-activity._udp");
+  priv->activity_browser = ga_service_browser_new
+      (SALUT_DNSSD_OLPC_ACTIVITY);
 #endif
 
   return ret;
@@ -735,7 +868,7 @@ salut_contact_manager_new(SalutConnection *connection) {
 
 gboolean
 salut_contact_manager_start(SalutContactManager *mgr,
-                            SalutAvahiClient *client, GError **error) {
+                            GaClient *client, GError **error) {
   SalutContactManagerPrivate *priv = SALUT_CONTACT_MANAGER_GET_PRIVATE (mgr);
 
   g_assert(priv->client == NULL);
@@ -750,7 +883,7 @@ salut_contact_manager_start(SalutContactManager *mgr,
   g_signal_connect(priv->presence_browser, "failure",
                    G_CALLBACK(browser_failed), mgr);
 
-  if (!salut_avahi_service_browser_attach(priv->presence_browser,
+  if (!ga_service_browser_attach(priv->presence_browser,
         priv->client, error))
     {
       return FALSE;
@@ -764,7 +897,7 @@ salut_contact_manager_start(SalutContactManager *mgr,
   g_signal_connect (priv->activity_browser, "failure",
       G_CALLBACK (browser_failed), mgr);
 
-  if (!salut_avahi_service_browser_attach(priv->activity_browser,
+  if (!ga_service_browser_attach(priv->activity_browser,
         priv->client, error))
     {
       return FALSE;
