@@ -22,9 +22,9 @@
 #define __SALUT_CONTACT_MANAGER_H__
 
 #include <glib-object.h>
+#include <avahi-gobject/ga-client.h>
 
 #include "salut-connection.h"
-#include "salut-avahi-client.h"
 #include "salut-contact.h"
 
 G_BEGIN_DECLS
@@ -61,7 +61,7 @@ SalutContactManager *
 salut_contact_manager_new(SalutConnection *connection);
 
 gboolean salut_contact_manager_start(SalutContactManager *mgr,
-                                     SalutAvahiClient *client,
+                                     GaClient *client,
                                      GError **error);
 
 
@@ -75,7 +75,17 @@ salut_contact_manager_find_contacts_by_address(SalutContactManager *mgr,
 #ifdef ENABLE_OLPC
 gboolean salut_contact_manager_merge_olpc_activity_properties
   (SalutContactManager *self, TpHandle handle, const gchar **color,
-   const gchar **name, const gchar **type);
+   const gchar **name, const gchar **type, const gchar **tags,
+   gboolean *is_private);
+
+void
+salut_contact_manager_add_invited_olpc_activity (SalutContactManager *self,
+    SalutContact *invitor, TpHandle room, const gchar *activity_id,
+    const gchar *color, const gchar *activity_name,
+    const gchar *activity_type, const gchar *tags);
 #endif
 
+SalutContact * salut_contact_manager_ensure_contact (SalutContactManager *mgr,
+    const gchar *name);
+
 #endif /* #ifndef __SALUT_CONTACT_MANAGER_H__*/
