@@ -26,10 +26,8 @@
 #include "salut-presence.h"
 #include "salut-presence-enumtypes.h"
 
-
-#include "salut-avahi-enums.h"
-#include "salut-avahi-service-resolver.h"
-#include "salut-avahi-record-browser.h"
+#include <avahi-gobject/ga-service-resolver.h>
+#include <avahi-gobject/ga-record-browser.h>
 #include <avahi-common/address.h>
 #include <avahi-common/defs.h>
 #include <avahi-common/malloc.h>
@@ -64,6 +62,7 @@
   g_free (_type);                                                           \
 } G_STMT_END
 
+#define PRESENCE_TIMEOUT (1200 * 1000)
 
 G_DEFINE_TYPE(SalutContact, salut_contact, G_TYPE_OBJECT)
 
@@ -91,6 +90,9 @@ static void
 salut_contact_avatar_request_flush(SalutContact *contact,
     guint8 *data, gsize size);
 
+static void
+salut_contact_retrieve_avatar (SalutContact *contact);
+
 #ifdef ENABLE_OLPC
 typedef struct {
   char *activity_id;
@@ -124,16 +126,20 @@ typedef struct _SalutContactPrivate SalutContactPrivate;
 
 struct _SalutContactPrivate
 {
+  SalutConnection *connection;
   gboolean dispose_has_run;
   gchar *alias;
-  SalutAvahiClient *client;
+  GaClient *client;
   GList *resolvers;
   gboolean found;
-  SalutAvahiRecordBrowser *record_browser;
+  GaRecordBrowser *record_browser;
   GList *avatar_requests;
-  TpHandleRepoIface *room_repo;
+  guint presence_resolver_failed_timer;
 #ifdef ENABLE_OLPC
-  GHashTable *olpc_activities;
+  /* mDNS instance name -> SalutContactActivity */
+  GHashTable *olpc_announced_activities;
+  /* activity id -> SalutContactActivity */
+  GHashTable *olpc_private_activities;
 #endif
 };
 
@@ -155,8 +161,12 @@ salut_contact_init (SalutContact *obj)
   obj->olpc_color = NULL;
   obj->olpc_cur_act = NULL;
   obj->olpc_cur_act_room = 0;
-  priv->olpc_activities = g_hash_table_new_full (g_str_hash, g_str_equal,
-      (GDestroyNotify) g_free, (GDestroyNotify) activity_free);
+  obj->olpc_ip4 = NULL;
+  obj->olpc_ip6 = NULL;
+  priv->olpc_announced_activities = g_hash_table_new_full (g_str_hash,
+      g_str_equal, (GDestroyNotify) g_free, (GDestroyNotify) activity_free);
+  priv->olpc_private_activities = g_hash_table_new_full (g_str_hash,
+      g_str_equal, (GDestroyNotify) g_free, (GDestroyNotify) activity_free);
 #endif
   priv->client = NULL;
   priv->resolvers = NULL;
@@ -206,10 +216,10 @@ salut_contact_class_init (SalutContactClass *salut_contact_class)
   signals[ACTIVITY_CHANGE] = g_signal_new("activity-change",
       G_OBJECT_CLASS_TYPE (salut_contact_class),
       G_SIGNAL_RUN_LAST, 0, NULL, NULL,
-      salut_signals_marshal_VOID__STRING_UINT_STRING_STRING_STRING_STRING,
-      G_TYPE_NONE,
-      6, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING, G_TYPE_STRING,
-      G_TYPE_STRING, G_TYPE_STRING);
+      salut_signals_marshal_VOID__STRING_UINT_STRING_STRING_STRING_STRING_STRING,
+      G_TYPE_NONE, 7,
+      G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING, G_TYPE_STRING,
+      G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);
 #endif
 }
 
@@ -218,6 +228,8 @@ salut_contact_dispose (GObject *object)
 {
   SalutContact *self = SALUT_CONTACT (object);
   SalutContactPrivate *priv = SALUT_CONTACT_GET_PRIVATE (self);
+  TpHandleRepoIface *contact_repo = tp_base_connection_get_handles
+      ((TpBaseConnection *) priv->connection, TP_HANDLE_TYPE_CONTACT);
 
   DEBUG_CONTACT (self, "Disposing contact");
 
@@ -229,15 +241,17 @@ salut_contact_dispose (GObject *object)
 #ifdef ENABLE_OLPC
   if (self->olpc_cur_act_room != 0)
     {
-      tp_handle_unref (priv->room_repo, self->olpc_cur_act_room);
+      TpHandleRepoIface *room_repo = tp_base_connection_get_handles
+          ((TpBaseConnection *) priv->connection, TP_HANDLE_TYPE_ROOM);
+
+      tp_handle_unref (room_repo, self->olpc_cur_act_room);
       self->olpc_cur_act_room = 0;
     }
 
-  g_hash_table_destroy (priv->olpc_activities);
+  g_hash_table_destroy (priv->olpc_announced_activities);
+  g_hash_table_destroy (priv->olpc_private_activities);
 #endif
 
-  priv->room_repo = NULL;
-
   salut_contact_avatar_request_flush(self, NULL, 0);
 
   /* release any references held by the object here */
@@ -249,6 +263,15 @@ salut_contact_dispose (GObject *object)
   g_list_free(priv->resolvers);
   priv->resolvers = NULL;
 
+  if (self->handle != 0)
+    tp_handle_unref (contact_repo, self->handle);
+
+  if (priv->presence_resolver_failed_timer > 0)
+    {
+      g_source_remove (priv->presence_resolver_failed_timer);
+      priv->presence_resolver_failed_timer = 0;
+    }
+
   if (G_OBJECT_CLASS (salut_contact_parent_class)->dispose)
     G_OBJECT_CLASS (salut_contact_parent_class)->dispose (object);
 }
@@ -272,6 +295,8 @@ salut_contact_finalize (GObject *object) {
     }
   g_free (self->olpc_color);
   g_free (self->olpc_cur_act);
+  g_free (self->olpc_ip4);
+  g_free (self->olpc_ip6);
 #endif
 
   G_OBJECT_CLASS (salut_contact_parent_class)->finalize (object);
@@ -321,7 +346,7 @@ compare_resolver(gconstpointer a, gconstpointer b) {
   return result;
 }
 
-static SalutAvahiServiceResolver *
+static GaServiceResolver *
 find_resolver(SalutContact *contact,
               AvahiIfIndex interface, AvahiProtocol protocol,
               const gchar *name, const gchar *type, const gchar *domain) {
@@ -334,16 +359,17 @@ find_resolver(SalutContact *contact,
   info.type = type;
   info.domain = domain;
   ret = g_list_find_custom(priv->resolvers, &info, compare_resolver);
-  return ret ? SALUT_AVAHI_SERVICE_RESOLVER(ret->data) : NULL;
+  return ret ? GA_SERVICE_RESOLVER(ret->data) : NULL;
 }
 
 SalutContact *
-salut_contact_new(SalutAvahiClient *client,
-                  TpHandleRepoIface *room_repo,
+salut_contact_new(GaClient *client,
+                  SalutConnection *connection,
                   const gchar *name)
 {
   SalutContact *ret;
   SalutContactPrivate *priv;
+  TpHandleRepoIface *contact_repo;
 
   ret = g_object_new(SALUT_TYPE_CONTACT, NULL);
   priv = SALUT_CONTACT_GET_PRIVATE (ret);
@@ -352,7 +378,12 @@ salut_contact_new(SalutAvahiClient *client,
 
   g_object_ref(client);
   priv->client = client;
-  priv->room_repo = room_repo;
+  priv->connection = connection;
+
+  contact_repo = tp_base_connection_get_handles
+      ((TpBaseConnection *) priv->connection, TP_HANDLE_TYPE_CONTACT);
+
+  ret->handle = tp_handle_ensure (contact_repo, name, NULL, NULL);
 
   return ret;
 }
@@ -380,8 +411,19 @@ update_alias(SalutContact *self, const gchar *new, gboolean *valid) {
 
 static void
 purge_cached_avatar(SalutContact *self, const gchar *token) {
-  g_free(self->avatar_token);
-  self->avatar_token = g_strdup(token);
+  SalutContactPrivate *priv = SALUT_CONTACT_GET_PRIVATE (self);
+
+  g_free (self->avatar_token);
+  self->avatar_token = g_strdup (token);
+
+  /* the avatar token has changed, restart retrieving the avatar just to be
+   * sure */
+  if (priv->record_browser != NULL)
+    {
+      g_object_unref (priv->record_browser);
+      priv->record_browser = NULL;
+      salut_contact_retrieve_avatar (self);
+    }
 }
 
 #ifdef ENABLE_OLPC
@@ -411,14 +453,16 @@ salut_contact_foreach_olpc_activity (SalutContact *self,
 
   DEBUG ("called");
 
-  g_hash_table_foreach (priv->olpc_activities, foreach_olpc_activity,
+  g_hash_table_foreach (priv->olpc_announced_activities, foreach_olpc_activity,
+      &ctx);
+  g_hash_table_foreach (priv->olpc_private_activities, foreach_olpc_activity,
       &ctx);
 
   DEBUG ("end");
 }
 
 static void
-activity_resolved_cb (SalutAvahiServiceResolver *resolver,
+activity_resolved_cb (GaServiceResolver *resolver,
                       AvahiIfIndex interface,
                       AvahiProtocol protocol,
                       gchar *name,
@@ -433,12 +477,15 @@ activity_resolved_cb (SalutAvahiServiceResolver *resolver,
 {
   SalutContact *self = SALUT_CONTACT (userdata);
   SalutContactPrivate *priv = SALUT_CONTACT_GET_PRIVATE (self);
+  TpHandleRepoIface *room_repo = tp_base_connection_get_handles
+      ((TpBaseConnection *) priv->connection, TP_HANDLE_TYPE_ROOM);
   AvahiStringList *t;
   TpHandle room_handle = 0;
   char *activity_id = NULL;
   char *color = NULL;
   char *activity_name = NULL;
   char *activity_type = NULL;
+  char *tags = NULL;
   SalutContactActivity *activity;
 
   DEBUG ("called: \"%s\".%s. on %s port %u", name, domain, host_name, port);
@@ -458,16 +505,20 @@ activity_resolved_cb (SalutAvahiServiceResolver *resolver,
       avahi_free (txtvers);
     }
 
-  activity = g_hash_table_lookup (priv->olpc_activities, name);
+  activity = g_hash_table_lookup (priv->olpc_announced_activities, name);
   if (activity == NULL)
     {
-      activity = activity_new (priv->room_repo);
-      g_hash_table_insert (priv->olpc_activities, g_strdup (name), activity);
+      activity = activity_new (room_repo);
+      g_hash_table_insert (priv->olpc_announced_activities, g_strdup (name),
+          activity);
     }
 
   if ((t = avahi_string_list_find (txt, "activity-id")) != NULL)
     {
       avahi_string_list_get_pair (t, NULL, &activity_id, NULL);
+
+      /* remove from the private activity list if needed */
+      g_hash_table_remove (priv->olpc_private_activities, activity_id);
     }
 
   if ((t = avahi_string_list_find (txt, "room")) != NULL)
@@ -475,7 +526,7 @@ activity_resolved_cb (SalutAvahiServiceResolver *resolver,
       char *room_name;
 
       avahi_string_list_get_pair (t, NULL, &room_name, NULL);
-      room_handle = tp_handle_ensure (priv->room_repo, room_name, NULL, NULL);
+      room_handle = tp_handle_ensure (room_repo, room_name, NULL, NULL);
       avahi_free (room_name);
     }
 
@@ -494,11 +545,16 @@ activity_resolved_cb (SalutAvahiServiceResolver *resolver,
       avahi_string_list_get_pair (t, NULL, &activity_type, NULL);
     }
 
+  if ((t = avahi_string_list_find (txt, "tags")) != NULL)
+    {
+      avahi_string_list_get_pair (t, NULL, &tags, NULL);
+    }
+
   if (room_handle == 0 || activity_id == NULL || activity_id[0] == '\0')
     {
       DEBUG ("Activity ID %s, room handle %u: removing from hash",
           activity_id ? activity_id : "<NULL>", room_handle);
-      g_hash_table_remove (priv->olpc_activities, name);
+      g_hash_table_remove (priv->olpc_announced_activities, name);
 
       g_signal_emit (self, signals[CONTACT_CHANGE], 0,
           SALUT_CONTACT_OLPC_ACTIVITIES);
@@ -517,7 +573,7 @@ activity_resolved_cb (SalutAvahiServiceResolver *resolver,
       activity_id = NULL;
 
       if (activity->room != 0)
-        tp_handle_unref (priv->room_repo, activity->room);
+        tp_handle_unref (room_repo, activity->room);
       activity->room = room_handle;
       room_handle = 0;
 
@@ -527,22 +583,103 @@ activity_resolved_cb (SalutAvahiServiceResolver *resolver,
 
   g_signal_emit (self, signals[ACTIVITY_CHANGE], 0,
       name, activity->room, activity->activity_id,
-      color, activity_name, activity_type);
+      color, activity_name, activity_type, tags);
 
   if (room_handle != 0)
-    tp_handle_unref (priv->room_repo, room_handle);
+    tp_handle_unref (room_repo, room_handle);
   avahi_free (activity_id);
   avahi_free (activity_type);
   avahi_free (activity_name);
   avahi_free (color);
+  avahi_free (tags);
 }
+
+static gboolean
+activity_has_id (gpointer key,
+                 gpointer value,
+                 gpointer user_data)
+{
+  SalutContactActivity *activity = value;
+  const gchar *activity_id = user_data;
+
+  return (!tp_strdiff (activity->activity_id, activity_id));
+}
+
+void
+salut_contact_takes_part_in_olpc_activity (SalutContact *self,
+                                           TpHandle room,
+                                           const gchar *activity_id)
+{
+  SalutContactPrivate *priv = SALUT_CONTACT_GET_PRIVATE (self);
+  SalutContactActivity *activity;
+  TpHandleRepoIface *room_repo = tp_base_connection_get_handles
+      ((TpBaseConnection *) priv->connection, TP_HANDLE_TYPE_ROOM);
+
+  activity = g_hash_table_lookup (priv->olpc_private_activities, activity_id);
+  if (activity != NULL)
+    {
+      DEBUG_CONTACT (self,
+          "activity is already in contact's private activities list");
+      return;
+    }
+
+  /* TODO: maybe we should add a hash that maps activity ID => activity for
+   * announced activities */
+  activity = g_hash_table_find (priv->olpc_announced_activities,
+      activity_has_id, (gchar *) activity_id);
+  if (activity != NULL)
+    {
+      DEBUG_CONTACT (self,
+          "activity is already in contact's announced activities list");
+      return;
+    }
+
+  DEBUG_CONTACT (self, "added in activity %s as he just invited us",
+      activity_id);
+
+  activity = activity_new (room_repo);
+  g_hash_table_insert (priv->olpc_private_activities,
+      g_strdup (activity_id), activity);
+
+  activity->activity_id = g_strdup (activity_id);
+  activity->room = room;
+  tp_handle_ref (room_repo, activity->room);
+
+  g_signal_emit (self, signals[CONTACT_CHANGE], 0,
+      SALUT_CONTACT_OLPC_ACTIVITIES);
+}
+
+void
+salut_contact_left_private_activity (SalutContact *self,
+                                     TpHandle room,
+                                     const gchar *activity_id)
+{
+  SalutContactPrivate *priv = SALUT_CONTACT_GET_PRIVATE (self);
+  SalutContactActivity *activity;
+
+  activity = g_hash_table_lookup (priv->olpc_private_activities, activity_id);
+  if (activity == NULL)
+    return;
+
+  if (activity->room != room)
+    {
+      DEBUG ("bad room handle: %d", room);
+      return;
+    }
+
+  g_hash_table_remove (priv->olpc_private_activities, activity_id);
+
+  g_signal_emit (self, signals[CONTACT_CHANGE], 0,
+      SALUT_CONTACT_OLPC_ACTIVITIES);
+}
+
 #endif
 
 static void
-contact_resolved_cb(SalutAvahiServiceResolver *resolver,
+contact_resolved_cb(GaServiceResolver *resolver,
                     AvahiIfIndex interface, AvahiProtocol protocol,
                     gchar *name, gchar *type, gchar *domain, gchar *host_name,
-                    AvahiAddress *a, gint port,
+                    AvahiAddress *address, gint port,
                     AvahiStringList *txt, AvahiLookupResultFlags flags,
                     gpointer userdata) {
   SalutContact *self = SALUT_CONTACT (userdata);
@@ -550,9 +687,20 @@ contact_resolved_cb(SalutAvahiServiceResolver *resolver,
   AvahiStringList *t;
   gint changes = 0;
   gboolean alias_seen = FALSE;
+#ifdef ENABLE_OLPC
+  TpHandleRepoIface *room_repo = tp_base_connection_get_handles
+    ((TpBaseConnection *) priv->connection, TP_HANDLE_TYPE_ROOM);
+#endif
 
   DEBUG_RESOLVER (self, resolver, "contact %s resolved", self->name);
 
+  if (priv->presence_resolver_failed_timer != 0)
+    {
+      DEBUG_CONTACT (self, "remove presence resolver timer");
+      g_source_remove (priv->presence_resolver_failed_timer);
+      priv->presence_resolver_failed_timer = 0;
+    }
+
 #define SET_CHANGE(x) changes |= x
 
   if ((t = avahi_string_list_find(txt, "status")) != NULL) {
@@ -728,7 +876,7 @@ contact_resolved_cb(SalutAvahiServiceResolver *resolver,
             }
           else
             {
-              room_handle = tp_handle_ensure (priv->room_repo, room_value,
+              room_handle = tp_handle_ensure (room_repo, room_value,
                   NULL, NULL);
             }
           avahi_free (room_value);
@@ -749,7 +897,7 @@ contact_resolved_cb(SalutAvahiServiceResolver *resolver,
             {
               g_free (self->olpc_cur_act);
               if (self->olpc_cur_act_room != 0)
-                tp_handle_unref (priv->room_repo, self->olpc_cur_act_room);
+                tp_handle_unref (room_repo, self->olpc_cur_act_room);
               self->olpc_cur_act = NULL;
               self->olpc_cur_act_room = 0;
               SET_CHANGE (SALUT_CONTACT_OLPC_CURRENT_ACTIVITY);
@@ -764,14 +912,14 @@ contact_resolved_cb(SalutAvahiServiceResolver *resolver,
             {
               g_free (self->olpc_cur_act);
               if (self->olpc_cur_act_room != 0)
-                tp_handle_unref (priv->room_repo, self->olpc_cur_act_room);
+                tp_handle_unref (room_repo, self->olpc_cur_act_room);
               self->olpc_cur_act_room = room_handle;
               self->olpc_cur_act = g_strdup (act_value);
               SET_CHANGE (SALUT_CONTACT_OLPC_CURRENT_ACTIVITY);
             }
           else
             {
-              tp_handle_unref (priv->room_repo, room_handle);
+              tp_handle_unref (room_repo, room_handle);
             }
           avahi_free (act_value);
         }
@@ -818,6 +966,36 @@ contact_resolved_cb(SalutAvahiServiceResolver *resolver,
           g_array_free (new_key, TRUE);
         }
     }
+
+  if (address != NULL)
+    {
+      gchar* saddr = g_malloc0 (AVAHI_ADDRESS_STR_MAX);
+
+      if (avahi_address_snprint (saddr, AVAHI_ADDRESS_STR_MAX, address))
+        {
+          switch (address->proto)
+            {
+              case AVAHI_PROTO_INET:
+                g_free (self->olpc_ip4);
+                self->olpc_ip4 = saddr;
+                SET_CHANGE (SALUT_CONTACT_OLPC_PROPERTIES);
+                break;
+              case AVAHI_PROTO_INET6:
+                g_free (self->olpc_ip6);
+                self->olpc_ip6 = saddr;
+                SET_CHANGE (SALUT_CONTACT_OLPC_PROPERTIES);
+                break;
+              default:
+                g_free (saddr);
+                break;
+            }
+        }
+      else
+        {
+          g_free (saddr);
+        }
+    }
+
 #endif
 
   if (changes != 0) {
@@ -843,7 +1021,7 @@ contact_lost(SalutContact *contact) {
 
 static void
 contact_drop_resolver (SalutContact *self,
-                       SalutAvahiServiceResolver *resolver)
+                       GaServiceResolver *resolver)
 {
   SalutContactPrivate *priv = SALUT_CONTACT_GET_PRIVATE (self);
   gint resolvers_left;
@@ -855,10 +1033,10 @@ contact_drop_resolver (SalutContact *self,
       "type", &type,
       "name", &name,
       NULL);
-  if (!tp_strdiff (type, "_olpc-activity._udp"))
+  if (!tp_strdiff (type, SALUT_DNSSD_OLPC_ACTIVITY))
     {
       /* one of their activities has fallen off */
-      g_hash_table_remove (priv->olpc_activities, name);
+      g_hash_table_remove (priv->olpc_announced_activities, name);
       g_signal_emit (self, signals[CONTACT_CHANGE], 0,
           SALUT_CONTACT_OLPC_ACTIVITIES);
     }
@@ -868,7 +1046,7 @@ contact_drop_resolver (SalutContact *self,
 #endif
 
   /* FIXME: what we actually want to be counting is the number of _presence
-   * resolvers, ignoring the _olpc-activity ones. However, if someone's still
+   * resolvers, ignoring the OLPC activity ones. However, if someone's still
    * advertising activities, we can probably consider them to be online? */
   priv->resolvers = g_list_remove(priv->resolvers, resolver);
 
@@ -884,13 +1062,35 @@ contact_drop_resolver (SalutContact *self,
   }
 }
 
+static gboolean
+presence_resolver_failed_timeout (gpointer data)
+{
+  SalutContact *self = SALUT_CONTACT (data);
+  SalutContactPrivate *priv = SALUT_CONTACT_GET_PRIVATE (self);
+
+  DEBUG_CONTACT (self, "presence resolver timer expired. Remove contact");
+  priv->presence_resolver_failed_timer = 0;
+  contact_lost (self);
+
+  return FALSE;
+}
+
 static void
-contact_failed_cb(SalutAvahiServiceResolver *resolver, GError *error,
+contact_failed_cb(GaServiceResolver *resolver, GError *error,
                    gpointer userdata) {
   SalutContact *self = SALUT_CONTACT (userdata);
+  SalutContactPrivate *priv = SALUT_CONTACT_GET_PRIVATE (self);
 
-  DEBUG_RESOLVER (self, resolver, "failed: %s. Ignoring, "
-    "presence might become stale", error->message);
+  if (priv->presence_resolver_failed_timer != 0)
+    /* There is already a timer running */
+    return;
+
+  DEBUG_RESOLVER (self, resolver, "failed: %s. Start presence resolver timer",
+      error->message);
+
+  priv->presence_resolver_failed_timer = g_timeout_add (
+      PRESENCE_TIMEOUT, presence_resolver_failed_timeout,
+      self);
 }
 
 void
@@ -899,20 +1099,21 @@ salut_contact_add_service(SalutContact *contact,
                           const char *name, const char *type,
                           const char *domain) {
   SalutContactPrivate *priv = SALUT_CONTACT_GET_PRIVATE (contact);
-  SalutAvahiServiceResolver *resolver;
+  GaServiceResolver *resolver;
   resolver = find_resolver(contact, interface, protocol, name, type, domain);
+  GError *error = NULL;
 
   if (resolver)
     return;
 
-  resolver = salut_avahi_service_resolver_new(interface,
+  resolver = ga_service_resolver_new(interface,
                                               protocol,
                                               name, type, domain,
                                               protocol, 0);
 
 #ifdef ENABLE_OLPC
   /* FIXME: better way to do this? */
-  if (!tp_strdiff (type, "_olpc-activity._udp"))
+  if (!tp_strdiff (type, SALUT_DNSSD_OLPC_ACTIVITY))
     {
       g_signal_connect(resolver, "found", G_CALLBACK(activity_resolved_cb),
           contact);
@@ -925,8 +1126,9 @@ salut_contact_add_service(SalutContact *contact,
     }
 
   g_signal_connect(resolver, "failure", G_CALLBACK(contact_failed_cb), contact);
-  if (!salut_avahi_service_resolver_attach(resolver, priv->client, NULL)) {
-    g_warning("Failed to attach resolver");
+  if (!ga_service_resolver_attach(resolver, priv->client, &error)) {
+    g_warning("Failed to attach resolver: %s", error->message);
+    g_error_free (error);
   }
 
   DEBUG_RESOLVER (contact, resolver, "added");
@@ -938,7 +1140,7 @@ salut_contact_remove_service(SalutContact *contact,
                           AvahiIfIndex interface, AvahiProtocol protocol,
                           const char *name, const char *type,
                           const char *domain) {
-  SalutAvahiServiceResolver *resolver;
+  GaServiceResolver *resolver;
   resolver =  find_resolver(contact, interface, protocol, name, type, domain);
 
   if (!resolver)
@@ -980,7 +1182,7 @@ _avahi_address_to_sockaddr(AvahiAddress *address, guint16 port,
 
 static void
 _add_address(gpointer data, gpointer user_data) {
-  SalutAvahiServiceResolver *resolver = SALUT_AVAHI_SERVICE_RESOLVER(data);
+  GaServiceResolver *resolver = GA_SERVICE_RESOLVER(data);
   GArray *addresses = (GArray *)user_data;
   salut_contact_address_t s_address;
   AvahiAddress address;
@@ -988,7 +1190,7 @@ _add_address(gpointer data, gpointer user_data) {
   AvahiIfIndex ifindex;
 
   g_object_get(resolver, "interface", &ifindex, NULL);
-  if (salut_avahi_service_resolver_get_address(resolver, &address, &port)) {
+  if (ga_service_resolver_get_address(resolver, &address, &port)) {
     _avahi_address_to_sockaddr(&address, port, ifindex, &(s_address.address));
     g_array_append_val(addresses, s_address);
   }
@@ -1009,7 +1211,7 @@ salut_contact_get_addresses(SalutContact *contact) {
 
 static gint
 _compare_address(gconstpointer a, gconstpointer b) {
-  SalutAvahiServiceResolver *resolver = SALUT_AVAHI_SERVICE_RESOLVER(a);
+  GaServiceResolver *resolver = GA_SERVICE_RESOLVER(a);
   struct sockaddr_storage addr_a;
   struct sockaddr_storage *addr_b = (struct sockaddr_storage *)b;
   AvahiIfIndex ifindex;
@@ -1017,7 +1219,7 @@ _compare_address(gconstpointer a, gconstpointer b) {
   uint16_t port;
 
   g_object_get(resolver, "interface", &ifindex, NULL);
-  if (!salut_avahi_service_resolver_get_address(resolver, &address, &port)) {
+  if (!ga_service_resolver_get_address(resolver, &address, &port)) {
     return -1;
   }
   _avahi_address_to_sockaddr(&address, port, ifindex, &addr_a);
@@ -1092,7 +1294,7 @@ salut_contact_avatar_request_flush(SalutContact *contact,
 }
 
 static void
-salut_contact_avatar_all_for_now(SalutAvahiRecordBrowser *browser,
+salut_contact_avatar_all_for_now(GaRecordBrowser *browser,
                                  gpointer user_data) {
   SalutContact *contact = SALUT_CONTACT(user_data);
 
@@ -1101,7 +1303,7 @@ salut_contact_avatar_all_for_now(SalutAvahiRecordBrowser *browser,
 }
 
 static void
-salut_contact_avatar_failure(SalutAvahiRecordBrowser *browser, GError *error,
+salut_contact_avatar_failure(GaRecordBrowser *browser, GError *error,
                              gpointer user_data) {
   SalutContact *contact = SALUT_CONTACT(user_data);
 
@@ -1111,7 +1313,7 @@ salut_contact_avatar_failure(SalutAvahiRecordBrowser *browser, GError *error,
 }
 
 static void
-salut_contact_avatar_found(SalutAvahiRecordBrowser *browser,
+salut_contact_avatar_found(GaRecordBrowser *browser,
                            AvahiIfIndex interface, AvahiProtocol protocol,
                            gchar *name, guint16 clazz, guint16 type,
                            guint8 *rdata, gsize rdata_size,
@@ -1136,8 +1338,8 @@ salut_contact_retrieve_avatar(SalutContact *contact) {
     return;
   }
 
-  name = g_strdup_printf("%s._presence._tcp.local", contact->name);
-  priv->record_browser = salut_avahi_record_browser_new(name, 0xA);
+  name = g_strdup_printf("%s." SALUT_DNSSD_PRESENCE ".local", contact->name);
+  priv->record_browser = ga_record_browser_new(name, 0xA);
   g_free(name);
 
   g_signal_connect(priv->record_browser, "new-record",
@@ -1147,7 +1349,7 @@ salut_contact_retrieve_avatar(SalutContact *contact) {
   g_signal_connect(priv->record_browser, "failure",
                    G_CALLBACK(salut_contact_avatar_failure), contact);
 
-  salut_avahi_record_browser_attach(priv->record_browser,
+  ga_record_browser_attach(priv->record_browser,
                                     priv->client, &error);
 }
 
