@@ -27,10 +27,11 @@
 
 #include <telepathy-glib/handle-repo.h>
 
-#include "salut-avahi-service-resolver.h"
-#include "salut-avahi-enums.h"
+#include <avahi-gobject/ga-service-resolver.h>
+#include <avahi-gobject/ga-enums.h>
 
 #include "salut-presence.h"
+#include "salut-connection.h"
 
 G_BEGIN_DECLS
 
@@ -58,11 +59,14 @@ struct _SalutContact {
     gchar *avatar_token;
     gchar *status_message;
     gchar *jid;
+    TpHandle handle;
 #ifdef ENABLE_OLPC
     GArray *olpc_key;
     gchar *olpc_color;
     gchar *olpc_cur_act;
     TpHandle olpc_cur_act_room;
+    gchar *olpc_ip4;
+    gchar *olpc_ip6;
 #endif /* ENABLE_OLPC */
 };
 
@@ -82,8 +86,8 @@ GType salut_contact_get_type(void);
 #define SALUT_CONTACT_GET_CLASS(obj) \
   (G_TYPE_INSTANCE_GET_CLASS ((obj), SALUT_TYPE_CONTACT, SalutContactClass))
 
-SalutContact *salut_contact_new(SalutAvahiClient *client,
-    TpHandleRepoIface *room_repo, const gchar *name);
+SalutContact *salut_contact_new (GaClient *client,
+    SalutConnection *conn, const gchar *name);
 
 void
 salut_contact_add_service(SalutContact *contact,
@@ -130,6 +134,13 @@ typedef void (*SalutContactOLPCActivityFunc)
 
 void salut_contact_foreach_olpc_activity (SalutContact *self,
     SalutContactOLPCActivityFunc foreach, gpointer user_data);
+
+void salut_contact_takes_part_in_olpc_activity (SalutContact *self,
+    TpHandle room, const gchar *activity_id);
+
+void salut_contact_left_private_activity (SalutContact *self, TpHandle room,
+    const gchar *activity_id);
+
 #endif
 
 G_END_DECLS
