@@ -36,7 +36,6 @@
 #include "salut-contact.h"
 #include "text-helper.h"
 #include "salut-xmpp-connection-manager.h"
-#include "namespaces.h"
 
 #include <gibber/gibber-linklocal-transport.h>
 #include <gibber/gibber-xmpp-connection.h>
@@ -829,8 +828,9 @@ salut_im_channel_is_text_message (GibberXmppStanza *stanza)
   if (type != GIBBER_STANZA_TYPE_MESSAGE)
     return FALSE;
 
-  if (gibber_xmpp_node_get_child_ns (stanza->node, "x", NS_LLMUC) != NULL)
-    /* discard MUC invite */
+  if (gibber_xmpp_node_get_child_ns (stanza->node, "invite",
+        GIBBER_TELEPATHY_NS_CLIQUE) != NULL)
+    /* discard Clique MUC invite */
     return FALSE;
 
   return TRUE;
