@@ -27,8 +27,12 @@
 #define DEBUG_FLAG DEBUG_MUC
 #include "debug.h"
 
+/* Maximum time to wait for others joining the group  */
+#define CONNECTED_TIMEOUT 60 * 1000
+
 #include "salut-muc-channel.h"
 
+#include <gibber/gibber-namespaces.h>
 #include <gibber/gibber-transport.h>
 
 #include <telepathy-glib/channel-iface.h>
@@ -38,12 +42,13 @@
 #include <telepathy-glib/util.h>
 
 #include <gibber/gibber-muc-connection.h>
+#include <avahi-gobject/ga-entry-group.h>
 
 #include "salut-connection.h"
+#include "salut-self.h"
 #include "salut-xmpp-connection-manager.h"
-#include "salut-avahi-entry-group.h"
+#include "salut-muc-manager.h"
 
-#include "namespaces.h"
 #include "text-helper.h"
 
 static void
@@ -59,6 +64,16 @@ G_DEFINE_TYPE_WITH_CODE(SalutMucChannel, salut_muc_channel, G_TYPE_OBJECT,
     G_IMPLEMENT_INTERFACE(TP_TYPE_SVC_CHANNEL_TYPE_TEXT, text_iface_init);
 )
 
+/* signal enum */
+enum
+{
+    READY,
+    JOIN_ERROR,
+    LAST_SIGNAL
+};
+
+static guint signals[LAST_SIGNAL] = {0};
+
 /* properties */
 enum
 {
@@ -70,6 +85,7 @@ enum
   PROP_CONNECTION,
   PROP_NAME,
   PROP_CLIENT,
+  PROP_CREATOR,
   PROP_XMPP_CONNECTION_MANAGER,
   LAST_PROPERTY
 };
@@ -83,12 +99,19 @@ struct _SalutMucChannelPrivate
   gchar *object_path;
   TpHandle handle;
   SalutConnection *connection;
+  SalutSelf *self;
   SalutXmppConnectionManager *xmpp_connection_manager;
   GibberMucConnection *muc_connection;
   gchar *muc_name;
-  SalutAvahiClient *client;
-  SalutAvahiEntryGroup *muc_group;
-  SalutAvahiEntryGroupService *service;
+  gboolean connected;
+  GaClient *client;
+  GaEntryGroup *muc_group;
+  GaEntryGroupService *service;
+  gboolean creator;
+  guint timeout;
+  /* (gchar *) -> (SalutContact *) */
+  GHashTable *senders;
+  SalutMucManager *muc_manager;
 };
 
 #define SALUT_MUC_CHANNEL_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), SALUT_TYPE_MUC_CHANNEL, SalutMucChannelPrivate))
@@ -142,6 +165,9 @@ salut_muc_channel_get_property (GObject    *object,
     case PROP_CLIENT:
       g_value_set_object (value, priv->client);
       break;
+    case PROP_CREATOR:
+      g_value_set_boolean (value, priv->creator);
+      break;
     case PROP_XMPP_CONNECTION_MANAGER:
       g_value_set_object (value, priv->xmpp_connection_manager);
       break;
@@ -192,6 +218,9 @@ salut_muc_channel_set_property (GObject     *object,
     case PROP_CLIENT:
       priv->client = g_value_get_object (value);
       break;
+    case PROP_CREATOR:
+      priv->creator = g_value_get_boolean (value);
+      break;
    case PROP_XMPP_CONNECTION_MANAGER:
       priv->xmpp_connection_manager = g_value_get_object (value);
       g_object_ref (priv->xmpp_connection_manager);
@@ -202,7 +231,62 @@ salut_muc_channel_set_property (GObject     *object,
   }
 }
 
+static void
+salut_muc_channel_add_self_to_members (SalutMucChannel *self)
+{
+  SalutMucChannelPrivate *priv = SALUT_MUC_CHANNEL_GET_PRIVATE (self);
+  TpBaseConnection *base_connection = TP_BASE_CONNECTION (priv->connection);
+  TpIntSet *empty;
+  TpIntSet *add;
 
+  priv->connected = TRUE;
+  g_signal_emit (self, signals[READY], 0);
+
+  if (priv->timeout != 0)
+    {
+      g_source_remove (priv->timeout);
+      priv->timeout = 0;
+    }
+
+  /* Now we are connected, move yourself to members */
+  empty = tp_intset_new ();
+  add = tp_intset_new ();
+  tp_intset_add (add, base_connection->self_handle);
+
+  tp_group_mixin_change_members (G_OBJECT (self),
+      "", add, empty, empty, empty, base_connection->self_handle,
+      TP_CHANNEL_GROUP_CHANGE_REASON_NONE);
+
+  tp_intset_destroy (empty);
+  tp_intset_destroy (add);
+}
+
+static gboolean
+connected_timeout_cb (gpointer user_data)
+{
+  SalutMucChannel *self = SALUT_MUC_CHANNEL (user_data);
+  SalutMucChannelPrivate *priv = SALUT_MUC_CHANNEL_GET_PRIVATE (self);
+
+  salut_muc_channel_add_self_to_members (self);
+  priv->timeout = 0;
+
+  return FALSE;
+}
+
+static void
+muc_connection_connected_cb (GibberMucConnection *connection,
+                             SalutMucChannel *self)
+{
+  SalutMucChannelPrivate *priv = SALUT_MUC_CHANNEL_GET_PRIVATE (self);
+
+  if (priv->creator)
+    salut_muc_channel_add_self_to_members (self);
+  else
+    priv->timeout = g_timeout_add (CONNECTED_TIMEOUT, connected_timeout_cb,
+        self);
+
+  salut_muc_channel_publish_service  (self);
+}
 
 static GObject *
 salut_muc_channel_constructor (GType type, guint n_props,
@@ -239,6 +323,21 @@ salut_muc_channel_constructor (GType type, guint n_props,
                                TP_CHANNEL_TEXT_MESSAGE_TYPE_ACTION,
                                G_MAXUINT);
 
+  g_object_get (priv->connection, "self", &(priv->self), NULL);
+  g_object_unref (priv->self);
+  g_assert (priv->self != NULL);
+
+  g_assert (priv->muc_connection != NULL);
+
+  priv->connected = FALSE;
+  g_signal_connect (priv->muc_connection, "connected",
+      G_CALLBACK (muc_connection_connected_cb), obj);
+
+  g_object_get (priv->connection,
+      "muc-manager", &(priv->muc_manager),
+      NULL);
+  g_assert (priv->muc_manager != NULL);
+
   /* Connect to the bus */
   bus = tp_get_bus ();
   dbus_g_connection_register_g_object(bus, priv->object_path, obj);
@@ -264,6 +363,9 @@ salut_muc_channel_init (SalutMucChannel *self)
   priv->muc_name = NULL;
   priv->client = NULL;
   priv->service = NULL;
+  priv->timeout = 0;
+  priv->senders = g_hash_table_new_full (g_str_hash, g_str_equal,
+      NULL, (GDestroyNotify) g_object_unref);
 }
 
 static void salut_muc_channel_dispose (GObject *object);
@@ -291,38 +393,46 @@ create_invitation (SalutMucChannel *self,
   TpHandleRepoIface *room_repo =
       tp_base_connection_get_handles(base_connection, TP_HANDLE_TYPE_ROOM);
   GibberXmppStanza *msg;
-  GibberXmppNode *x_node, *invite_node;
+  GibberXmppNode *invite_node;
 
   const gchar *name = tp_handle_inspect(contact_repo, handle);
 
-  msg = gibber_xmpp_stanza_new("message");
+  msg = gibber_xmpp_stanza_build (GIBBER_STANZA_TYPE_MESSAGE,
+      GIBBER_STANZA_SUB_TYPE_NORMAL,
+      priv->connection->name, name,
+      GIBBER_NODE, "body",
+        GIBBER_NODE_TEXT, "You got a Clique chatroom invitation",
+      GIBBER_NODE_END,
+      GIBBER_NODE, "invite",
+        GIBBER_NODE_ASSIGN_TO, &invite_node,
+        GIBBER_NODE_XMLNS, GIBBER_TELEPATHY_NS_CLIQUE,
+        GIBBER_NODE, "roomname",
+          GIBBER_NODE_TEXT, tp_handle_inspect (room_repo, priv->handle),
+        GIBBER_NODE_END,
+      GIBBER_NODE_END,
+      GIBBER_STANZA_END);
 
-  gibber_xmpp_node_set_attribute(msg->node, "from", priv->connection->name);
-  gibber_xmpp_node_set_attribute(msg->node, "to", name);
+  if (message != NULL && *message != '\0')
+    {
+      gibber_xmpp_node_add_child_with_content(invite_node, "reason",
+          message);
+    }
 
-  gibber_xmpp_node_add_child_with_content(msg->node, "body",
-                                         "You got a chatroom invitation");
-  x_node = gibber_xmpp_node_add_child_ns(msg->node, "x", NS_LLMUC);
+  g_hash_table_foreach (
+      (GHashTable *) gibber_muc_connection_get_parameters (
+          priv->muc_connection),
+      invitation_append_parameter, invite_node);
 
-  invite_node = gibber_xmpp_node_add_child(x_node, "invite");
-  gibber_xmpp_node_set_attribute(invite_node, "protocol",
-        gibber_muc_connection_get_protocol(priv->muc_connection));
-
-  if (message != NULL && *message != '\0') {
-    gibber_xmpp_node_add_child_with_content(invite_node, "reason", message);
-  }
-
-  gibber_xmpp_node_add_child_with_content(invite_node, "roomname",
-                       tp_handle_inspect(room_repo, priv->handle));
-  g_hash_table_foreach(
-    (GHashTable *)gibber_muc_connection_get_parameters(priv->muc_connection),
-    invitation_append_parameter, invite_node);
+#ifdef ENABLE_OLPC
+  salut_self_olpc_augment_invitation (priv->self, priv->handle, handle,
+      invite_node);
+#endif
 
   return msg;
 }
 
-static gboolean
-muc_channel_publish_service (SalutMucChannel *self)
+gboolean
+salut_muc_channel_publish_service (SalutMucChannel *self)
 {
   SalutMucChannelPrivate *priv = SALUT_MUC_CHANNEL_GET_PRIVATE (self);
   AvahiStringList *txt_record = NULL;
@@ -335,12 +445,22 @@ muc_channel_publish_service (SalutMucChannel *self)
   size_t dns_payload_length;
   AvahiAddress addr;
 
-  g_assert (priv->muc_group == NULL);
+  /* We are already announcing this muc group */
+  if (priv->muc_group != NULL)
+    return TRUE;
+
   g_assert (priv->service == NULL);
 
-  priv->muc_group = salut_avahi_entry_group_new ();
+  /* We didn't connect to this group just yet */
+  if (priv->muc_connection->state != GIBBER_MUC_CONNECTION_CONNECTED)
+    {
+      DEBUG ("Not yet connected to this muc, not announcing");
+      return TRUE;
+    }
 
-  if (!salut_avahi_entry_group_attach (priv->muc_group, priv->client, &error))
+  priv->muc_group = ga_entry_group_new ();
+
+  if (!ga_entry_group_attach (priv->muc_group, priv->client, &error))
     {
       DEBUG ("entry group attach failed: %s", error->message);
       goto publish_service_error;
@@ -382,10 +502,10 @@ muc_channel_publish_service (SalutMucChannel *self)
       goto publish_service_error;
     }
 
-  host = g_strdup_printf ("%s._salut-room._udp.local", priv->muc_name);
+  host = g_strdup_printf ("%s." SALUT_DNSSD_CLIQUE ".local", priv->muc_name);
 
   /* Add the record */
-  if (!salut_avahi_entry_group_add_record_full (priv->muc_group,
+  if (!ga_entry_group_add_record_full (priv->muc_group,
         AVAHI_IF_UNSPEC, addr.proto, 0,
         host, AVAHI_DNS_CLASS_IN, dns_type, AVAHI_DEFAULT_TTL_HOST_NAME,
         &(addr.data.data), dns_payload_length, &error))
@@ -400,16 +520,16 @@ muc_channel_publish_service (SalutMucChannel *self)
 
   /* We shouldn't add the service but manually create the SRV record so
    * we'll be able to allow multiple announcers */
-  priv->service = salut_avahi_entry_group_add_service_full_strlist (
+  priv->service = ga_entry_group_add_service_full_strlist (
       priv->muc_group, AVAHI_IF_UNSPEC, addr.proto, 0, priv->muc_name,
-      "_salut-room._udp", NULL, host, port, &error, txt_record);
+      SALUT_DNSSD_CLIQUE, NULL, host, port, &error, txt_record);
   if (priv->service == NULL)
     {
       DEBUG ("add service failed: %s", error->message);
       goto publish_service_error;
     }
 
-  if (!salut_avahi_entry_group_commit (priv->muc_group, &error))
+  if (!ga_entry_group_commit (priv->muc_group, &error))
     {
       DEBUG ("entry group commit failed: %s", error->message);
       goto publish_service_error;
@@ -530,56 +650,21 @@ xmpp_connection_manager_connection_failed_cb (SalutXmppConnectionManager *mgr,
       TP_CHANNEL_GROUP_CHANGE_REASON_ERROR);
 }
 
-static gboolean
-muc_channel_add_member (GObject *iface,
-                        TpHandle handle,
-                        const gchar *message,
-                        GError **error)
+gboolean
+salut_muc_channel_send_invitation (SalutMucChannel *self,
+                                   TpHandle handle,
+                                   const gchar *message,
+                                   GError **error)
 {
-  SalutMucChannel *self = SALUT_MUC_CHANNEL(iface);
   SalutMucChannelPrivate *priv = SALUT_MUC_CHANNEL_GET_PRIVATE (self);
-  TpBaseConnection *base_connection = TP_BASE_CONNECTION (priv->connection);
+  GibberXmppStanza *stanza;
   SalutContactManager *contact_manager = NULL;
   SalutContact *contact;
-  GibberXmppStanza *stanza;
-  GibberXmppConnection *connection = NULL;
   SalutXmppConnectionManagerRequestConnectionResult request_result;
   gboolean result;
   struct pending_connection_for_invite_data *data;
-  TpIntSet *empty, *remote_pending;
+  GibberXmppConnection *connection = NULL;
 
-  if (handle == base_connection->self_handle)
-    {
-      /* adding yourself, let's join the muc */
-      TpIntSet *empty;
-      TpIntSet *add;
-      gboolean ret = TRUE;
-
-      empty = tp_intset_new ();
-      add = tp_intset_new ();
-      tp_intset_add (add, handle);
-      /* Add to members */
-
-      if (salut_muc_channel_connect(self, NULL))
-        {
-          tp_group_mixin_change_members (G_OBJECT (self),
-              message, add, empty, empty, empty, base_connection->self_handle,
-              TP_CHANNEL_GROUP_CHANGE_REASON_INVITED);
-        }
-      else
-        {
-          g_set_error (error, TP_ERRORS, TP_ERROR_NETWORK_ERROR,
-              "Failed to connect to the group");
-          ret = FALSE;
-        }
-
-      tp_intset_destroy (empty);
-      tp_intset_destroy (add);
-      muc_channel_publish_service (self);
-      return ret;
-    }
-
-  /* Adding a contact, let's invite him */
   g_object_get (G_OBJECT (priv->connection), "contact-manager",
       &contact_manager, NULL);
   g_assert (contact_manager != NULL);
@@ -594,8 +679,6 @@ muc_channel_add_member (GObject *iface,
       return FALSE;
     }
 
-  DEBUG ("Trying to add handle %u to %s", handle ,priv->object_path);
-
   DEBUG ("request XMPP connection with contact %s", contact->name);
   request_result = salut_xmpp_connection_manager_request_connection (
       priv->xmpp_connection_manager, contact, &connection, error);
@@ -607,16 +690,6 @@ muc_channel_add_member (GObject *iface,
       return FALSE;
     }
 
-  /* Set the contact as remote pending */
-  empty = tp_intset_new ();
-  remote_pending = tp_intset_new ();
-  tp_intset_add (remote_pending, handle);
-  tp_group_mixin_change_members (G_OBJECT(self), "", empty, empty, empty,
-      remote_pending, base_connection->self_handle,
-      TP_CHANNEL_GROUP_CHANGE_REASON_INVITED);
-  tp_intset_destroy (empty);
-  tp_intset_destroy (remote_pending);
-
   stanza = create_invitation (self, handle, message);
 
   if (request_result ==
@@ -642,6 +715,76 @@ muc_channel_add_member (GObject *iface,
   return TRUE;
 }
 
+static gboolean
+muc_channel_add_member (GObject *iface,
+                        TpHandle handle,
+                        const gchar *message,
+                        GError **error)
+{
+  SalutMucChannel *self = SALUT_MUC_CHANNEL(iface);
+  SalutMucChannelPrivate *priv = SALUT_MUC_CHANNEL_GET_PRIVATE (self);
+  TpBaseConnection *base_connection = TP_BASE_CONNECTION (priv->connection);
+  TpIntSet *empty, *remote_pending;
+
+  if (handle == base_connection->self_handle)
+    {
+      /* adding yourself, let's join the muc */
+      TpIntSet *empty;
+      TpIntSet *add;
+      gboolean ret = TRUE;
+
+      if (tp_handle_set_is_member (self->group.remote_pending,
+          base_connection->self_handle))
+        {
+          /* Already in remote pending, no need to redo */
+          return TRUE;
+        }
+
+      empty = tp_intset_new ();
+      add = tp_intset_new ();
+      tp_intset_add (add, handle);
+      /* Add to members */
+
+      if (salut_muc_channel_connect(self, NULL))
+        {
+          /* We are considered as remote-pending while the muc connection
+           * is not connected */
+          tp_group_mixin_change_members (G_OBJECT (self),
+              message, empty, empty, empty, add, base_connection->self_handle,
+              TP_CHANNEL_GROUP_CHANGE_REASON_NONE);
+        }
+      else
+        {
+          g_set_error (error, TP_ERRORS, TP_ERROR_NETWORK_ERROR,
+              "Failed to connect to the group");
+          ret = FALSE;
+        }
+
+      tp_intset_destroy (empty);
+      tp_intset_destroy (add);
+      return ret;
+    }
+
+  /* Adding a contact, let's invite him */
+
+  DEBUG ("Trying to add handle %u to %s", handle ,priv->object_path);
+
+  if (!salut_muc_channel_send_invitation (self, handle, message, error))
+    return FALSE;
+
+  /* Set the contact as remote pending */
+  empty = tp_intset_new ();
+  remote_pending = tp_intset_new ();
+  tp_intset_add (remote_pending, handle);
+  tp_group_mixin_change_members (G_OBJECT(self), "", empty, empty, empty,
+      remote_pending, base_connection->self_handle,
+      TP_CHANNEL_GROUP_CHANGE_REASON_INVITED);
+  tp_intset_destroy (empty);
+  tp_intset_destroy (remote_pending);
+
+  return TRUE;
+}
+
 static void
 salut_muc_channel_class_init (SalutMucChannelClass *salut_muc_channel_class) {
   GObjectClass *object_class = G_OBJECT_CLASS (salut_muc_channel_class);
@@ -699,10 +842,10 @@ salut_muc_channel_class_init (SalutMucChannelClass *salut_muc_channel_class) {
 
   param_spec = g_param_spec_object (
       "client",
-      "SalutAvahiClient object",
+      "GaClient object",
       "Salut Avahi client used with the"
       " connection that owns this MUC channel",
-      SALUT_TYPE_AVAHI_CLIENT,
+      GA_TYPE_CLIENT,
       G_PARAM_CONSTRUCT_ONLY |
       G_PARAM_READWRITE |
       G_PARAM_STATIC_NICK |
@@ -722,6 +865,36 @@ salut_muc_channel_class_init (SalutMucChannelClass *salut_muc_channel_class) {
   g_object_class_install_property (object_class, PROP_XMPP_CONNECTION_MANAGER,
       param_spec);
 
+  param_spec = g_param_spec_boolean (
+      "creator",
+      "creator",
+      "Whether or not we created this muc",
+      FALSE,
+      G_PARAM_CONSTRUCT_ONLY |
+      G_PARAM_READWRITE |
+      G_PARAM_STATIC_NICK |
+      G_PARAM_STATIC_BLURB);
+  g_object_class_install_property (object_class,
+      PROP_CREATOR, param_spec);
+
+  signals[READY] = g_signal_new (
+        "ready",
+        G_OBJECT_CLASS_TYPE (salut_muc_channel_class),
+        G_SIGNAL_RUN_LAST,
+        0,
+        NULL, NULL,
+        g_cclosure_marshal_VOID__VOID,
+        G_TYPE_NONE, 0);
+
+  signals[JOIN_ERROR] = g_signal_new (
+        "join-error",
+        G_OBJECT_CLASS_TYPE (salut_muc_channel_class),
+        G_SIGNAL_RUN_LAST,
+        0,
+        NULL, NULL,
+        g_cclosure_marshal_VOID__POINTER,
+        G_TYPE_NONE, 1, G_TYPE_POINTER);
+
   tp_text_mixin_class_init(object_class,
                            G_STRUCT_OFFSET(SalutMucChannelClass, text_class));
 
@@ -761,6 +934,24 @@ salut_muc_channel_dispose (GObject *object)
       priv->muc_group = NULL;
     }
 
+  if (priv->timeout != 0)
+    {
+      g_source_remove (priv->timeout);
+      priv->timeout = 0;
+    }
+
+  if (priv->senders != NULL)
+    {
+      g_hash_table_destroy (priv->senders);
+      priv->senders = NULL;
+    }
+
+  if (priv->muc_manager != NULL)
+    {
+      g_object_unref (priv->muc_manager);
+      priv->muc_manager = NULL;
+    }
+
   /* release any references held by the object here */
   if (G_OBJECT_CLASS (salut_muc_channel_parent_class)->dispose)
     G_OBJECT_CLASS (salut_muc_channel_parent_class)->dispose (object);
@@ -783,7 +974,7 @@ salut_muc_channel_finalize (GObject *object)
 }
 
 gboolean
-salut_muc_channel_invited (SalutMucChannel *self, TpHandle invitor,
+salut_muc_channel_invited (SalutMucChannel *self, TpHandle inviter,
                           const gchar *stanza, GError **error) {
   SalutMucChannelPrivate *priv = SALUT_MUC_CHANNEL_GET_PRIVATE(self);
   TpBaseConnection *base_connection = TP_BASE_CONNECTION(priv->connection);
@@ -796,7 +987,7 @@ salut_muc_channel_invited (SalutMucChannel *self, TpHandle invitor,
   /* Got invited to this muc channel */
   DEBUG("Got an invitation to %s from %s",
     tp_handle_inspect(room_repo, priv->handle),
-    tp_handle_inspect(contact_repo, invitor)
+    tp_handle_inspect(contact_repo, inviter)
     );
 
   /* If we are already a member, no further actions are needed */
@@ -805,7 +996,7 @@ salut_muc_channel_invited (SalutMucChannel *self, TpHandle invitor,
     return TRUE;
   }
 
-  if (invitor == base_connection->self_handle) {
+  if (inviter == base_connection->self_handle) {
     /* Invited ourselves, go straight to members */
     gboolean r;
     GArray *members =  g_array_sized_new (FALSE, FALSE, sizeof(TpHandle), 1);
@@ -823,7 +1014,7 @@ salut_muc_channel_invited (SalutMucChannel *self, TpHandle invitor,
     tp_group_mixin_change_members(G_OBJECT(self), stanza,
                                   empty, empty,
                                   local_pending, empty,
-                                  invitor,
+                                  inviter,
                                   TP_CHANNEL_GROUP_CHANGE_REASON_INVITED);
     tp_intset_destroy(local_pending);
     tp_intset_destroy(empty);
@@ -851,23 +1042,85 @@ salut_muc_channel_send_stanza(SalutMucChannel *self, guint type,
 }
 
 static void
-salut_muc_channel_change_members(SalutMucChannel *self,
-                                 TpHandle from_handle,
-                                 gboolean joining) {
+salut_muc_channel_add_members (SalutMucChannel *self,
+                               GArray *members)
+{
+  SalutMucChannelPrivate *priv = SALUT_MUC_CHANNEL_GET_PRIVATE (self);
   TpIntSet *empty, *changes;
+  guint i;
+  SalutContactManager *contact_mgr;
 
-  empty = tp_intset_new();
-  changes = tp_intset_new();
-  tp_intset_add(changes, from_handle);
-  tp_group_mixin_change_members(G_OBJECT(self),
-                                "",
-                                joining ? changes : empty,
-                                joining ? empty : changes,
-                                empty, empty,
-                                from_handle,
-                                TP_CHANNEL_GROUP_CHANGE_REASON_NONE);
-  tp_intset_destroy(changes);
-  tp_intset_destroy(empty);
+  empty = tp_intset_new ();
+  changes = tp_intset_new ();
+
+  g_object_get (G_OBJECT (priv->connection), "contact-manager",
+      &contact_mgr, NULL);
+  g_assert (contact_mgr != NULL);
+
+  for (i = 0; i < members->len; i++)
+    {
+      gchar *sender = g_array_index (members, gchar *, i);
+      SalutContact *contact;
+
+      contact = salut_contact_manager_ensure_contact (contact_mgr, sender);
+
+      g_hash_table_insert (priv->senders, sender, contact);
+      tp_intset_add (changes, contact->handle);
+    }
+
+  tp_group_mixin_change_members (G_OBJECT(self),
+                                 "",
+                                 changes,
+                                 empty,
+                                 empty, empty,
+                                 0,
+                                 TP_CHANNEL_GROUP_CHANGE_REASON_NONE);
+  tp_intset_destroy (changes);
+  tp_intset_destroy (empty);
+  g_object_unref (contact_mgr);
+}
+
+static void
+salut_muc_channel_remove_members (SalutMucChannel *self,
+                                  GArray *members)
+{
+  SalutMucChannelPrivate *priv = SALUT_MUC_CHANNEL_GET_PRIVATE (self);
+  TpBaseConnection *base_connection =
+      (TpBaseConnection *) (priv->connection);
+  TpHandleRepoIface *contact_repo = tp_base_connection_get_handles
+      (base_connection, TP_HANDLE_TYPE_CONTACT);
+  TpIntSet *empty, *changes;
+  guint i;
+
+  empty = tp_intset_new ();
+  changes = tp_intset_new ();
+
+  for (i = 0; i < members->len; i++)
+    {
+      TpHandle handle;
+      gchar *sender = g_array_index (members, gchar *, i);
+
+      handle = tp_handle_lookup (contact_repo, sender, NULL, NULL);
+      if (handle == 0)
+        {
+          DEBUG ("Lost sender (%s), but unknown contact", sender);
+          continue;
+        }
+
+      g_hash_table_remove (priv->senders, sender);
+
+      tp_intset_add (changes, handle);
+    }
+
+  tp_group_mixin_change_members (G_OBJECT(self),
+                                 "",
+                                 empty,
+                                 changes,
+                                 empty, empty,
+                                 0,
+                                 TP_CHANNEL_GROUP_CHANGE_REASON_NONE);
+  tp_intset_destroy (changes);
+  tp_intset_destroy (empty);
 }
 
 static void
@@ -884,6 +1137,7 @@ salut_muc_channel_received_stanza(GibberMucConnection *conn,
   const gchar *from, *to, *body, *body_offset;
   TpChannelTextMessageType msgtype;
   TpHandle from_handle;
+  GibberXmppNode *tubes_node;
 
   to = gibber_xmpp_node_get_attribute(stanza->node, "to");
   if (strcmp(to, priv->muc_name)) {
@@ -891,6 +1145,35 @@ salut_muc_channel_received_stanza(GibberMucConnection *conn,
     return;
   }
 
+  from_handle = tp_handle_lookup(contact_repo, sender, NULL, NULL);
+  if (from_handle == 0) {
+    /* FIXME, unknown contact.. Need some way to handle this safely,
+     * just adding the contact is somewhat scary */
+    DEBUG("Got stanza from unknown contact, discarding");
+    return;
+  }
+
+#ifdef ENABLE_OLPC
+  if (salut_connection_olpc_observe_muc_stanza (priv->connection, priv->handle,
+        from_handle, stanza))
+    return;
+#endif
+
+  tubes_node = gibber_xmpp_node_get_child_ns (stanza->node, "tubes",
+      GIBBER_TELEPATHY_NS_TUBES);
+  if (tubes_node != NULL)
+    {
+      SalutTubesChannel *tubes_chan;
+
+      tubes_chan = salut_muc_manager_ensure_tubes_channel (priv->muc_manager,
+          priv->handle);
+      g_assert (tubes_chan != NULL);
+
+      tubes_message_received (tubes_chan, sender, stanza);
+
+      g_object_unref (tubes_chan);
+    }
+
   if (!text_helper_parse_incoming_message(stanza, &from, &msgtype,
                                           &body, &body_offset)) {
     DEBUG("Couldn't parse stanza");
@@ -902,82 +1185,75 @@ salut_muc_channel_received_stanza(GibberMucConnection *conn,
     return;
   }
 
-  from_handle = tp_handle_lookup(contact_repo, sender, NULL, NULL);
-  if (from_handle == 0) {
-    /* FIXME, unknown contact.. Need some way to handle this safely,
-     * just adding the contact is somewhat scary */
-    DEBUG("Got stanza from unknown contact, discarding");
-    return;
-  }
-
   /* FIXME validate the from and the to */
   tp_text_mixin_receive(G_OBJECT(self), msgtype, from_handle,
       time(NULL), body_offset);
 }
 
 static void
-salut_muc_channel_new_sender(GibberMucConnection *connection, gchar *sender,
-                             gpointer user_data) {
-  SalutMucChannel *self = SALUT_MUC_CHANNEL(user_data);
-  SalutMucChannelPrivate *priv = SALUT_MUC_CHANNEL_GET_PRIVATE(self);
-  TpBaseConnection *base_connection = TP_BASE_CONNECTION(priv->connection);
-  TpHandleRepoIface *contact_repo =
-      tp_base_connection_get_handles(base_connection, TP_HANDLE_TYPE_CONTACT);
-  TpHandle handle;
+salut_muc_channel_new_senders (GibberMucConnection *connection,
+                               GArray *senders,
+                               gpointer user_data)
+{
+  SalutMucChannel *self = SALUT_MUC_CHANNEL (user_data);
+  SalutMucChannelPrivate *priv = SALUT_MUC_CHANNEL_GET_PRIVATE (self);
+  TpBaseConnection *base_connection = TP_BASE_CONNECTION (priv->connection);
 
-  handle = tp_handle_lookup(contact_repo, sender, NULL, NULL);
-  /* FIXME what to do with invalid handles */
-  if (handle == 0) {
-    DEBUG("New sender, but unknown contact");
-    return;
-  }
-
-  salut_muc_channel_change_members(self, handle, TRUE);
+  salut_muc_channel_add_members (self, senders);
+  if (!tp_handle_set_is_member (self->group.members,
+      base_connection->self_handle))
+    salut_muc_channel_add_self_to_members (self);
 }
 
 static void
-salut_muc_channel_lost_sender(GibberMucConnection *connection, gchar *sender,
-                             gpointer user_data) {
+salut_muc_channel_lost_senders(GibberMucConnection *connection,
+    GArray *senders, gpointer user_data) {
   SalutMucChannel *self = SALUT_MUC_CHANNEL(user_data);
-  SalutMucChannelPrivate *priv = SALUT_MUC_CHANNEL_GET_PRIVATE(self);
-  TpBaseConnection *base_connection = TP_BASE_CONNECTION(priv->connection);
-  TpHandleRepoIface *contact_repo =
-      tp_base_connection_get_handles(base_connection, TP_HANDLE_TYPE_CONTACT);
-  TpHandle handle;
 
-  handle = tp_handle_lookup(contact_repo, sender, NULL, NULL);
-  /* FIXME what to do with invalid handles */
-  if (handle == 0) {
-    DEBUG("Lost sender, but unknown contact");
-    return;
-  }
-
-  salut_muc_channel_change_members(self, handle, FALSE);
+  salut_muc_channel_remove_members (self, senders);
 }
 
 static gboolean
-salut_muc_channel_connect(SalutMucChannel *channel, GError **error) {
-  SalutMucChannelPrivate *priv = SALUT_MUC_CHANNEL_GET_PRIVATE(channel);
+salut_muc_channel_connect (SalutMucChannel *channel,
+                           GError **error)
+{
+  SalutMucChannelPrivate *priv = SALUT_MUC_CHANNEL_GET_PRIVATE (channel);
 
-  g_signal_connect(priv->muc_connection, "received-stanza",
-                   G_CALLBACK(salut_muc_channel_received_stanza), channel);
+  g_signal_connect (priv->muc_connection, "received-stanza",
+      G_CALLBACK (salut_muc_channel_received_stanza), channel);
 
-  g_signal_connect(priv->muc_connection, "disconnected",
-                   G_CALLBACK(salut_muc_channel_disconnected), channel);
+  g_signal_connect (priv->muc_connection, "disconnected",
+      G_CALLBACK (salut_muc_channel_disconnected), channel);
 
-  g_signal_connect(priv->muc_connection, "new-sender",
-                   G_CALLBACK(salut_muc_channel_new_sender), channel);
+  g_signal_connect (priv->muc_connection, "new-senders",
+      G_CALLBACK (salut_muc_channel_new_senders), channel);
 
-  g_signal_connect(priv->muc_connection, "lost-sender",
-                   G_CALLBACK(salut_muc_channel_lost_sender), channel);
+  g_signal_connect (priv->muc_connection, "lost-senders",
+      G_CALLBACK (salut_muc_channel_lost_senders), channel);
 
-  return gibber_muc_connection_connect(priv->muc_connection, error);
+  return gibber_muc_connection_connect (priv->muc_connection, error);
 }
 
 static void
 salut_muc_channel_disconnected(GibberTransport *transport,
                                              gpointer user_data) {
   SalutMucChannel *self = SALUT_MUC_CHANNEL(user_data);
+  SalutMucChannelPrivate *priv = SALUT_MUC_CHANNEL_GET_PRIVATE (self);
+
+  if (priv->timeout != 0)
+    {
+      g_source_remove (priv->timeout);
+      priv->timeout = 0;
+    }
+
+  if (!priv->connected)
+    {
+      /* FIXME the disconnected signal should give us an error */
+      GError error = { TP_ERRORS, TP_ERROR_NETWORK_ERROR,
+          "can't join the muc" };
+      g_signal_emit (self, signals[JOIN_ERROR], 0, &error);
+    }
+
   tp_svc_channel_emit_closed(self);
 }
 
