@@ -25,20 +25,21 @@
 
 #include "salut-muc-manager.h"
 
+#include <avahi-gobject/ga-service-browser.h>
+
 #include <gibber/gibber-muc-connection.h>
+#include <gibber/gibber-namespaces.h>
+#include <gibber/gibber-xmpp-error.h>
 
 #include "salut-muc-channel.h"
 #include "salut-contact-manager.h"
-#include "salut-avahi-service-browser.h"
 #include "salut-tubes-channel.h"
+#include "salut-roomlist-channel.h"
 #include "salut-xmpp-connection-manager.h"
 
 #include <telepathy-glib/channel-factory-iface.h>
 #include <telepathy-glib/interfaces.h>
 
-#include "extensions/extensions.h"
-#include "namespaces.h"
-
 #define DEBUG_FLAG DEBUG_MUC
 #include "debug.h"
 
@@ -80,34 +81,54 @@ struct _SalutMucManagerPrivate
 
   /* GUINT_TO_POINTER (room_handle) => (SalutMucChannel *) */
   GHashTable *text_channels;
+#ifdef ENABLE_DBUS_TUBES
    /* GUINT_TO_POINTER(room_handle) => (SalutTubesChannel *) */
   GHashTable *tubes_channels;
+#endif
+  GSList *roomlist_channels;
 
   gboolean dispose_has_run;
-  SalutAvahiClient *client;
-  SalutAvahiServiceBrowser *browser;
-  /* room name => (SalutAvahiServiceResolver *) */
+  GaClient *client;
+  GaServiceBrowser *browser;
+  /* room name => (GaServiceResolver *) */
   GHashTable *room_resolvers;
 };
 
 #define SALUT_MUC_MANAGER_GET_PRIVATE(o)     (G_TYPE_INSTANCE_GET_PRIVATE ((o), SALUT_TYPE_MUC_MANAGER, SalutMucManagerPrivate))
 
+static void
+room_resolver_removed (gpointer data)
+{
+  GArray *arr = (GArray *) data;
+  int i;
+  for (i = 0; i < arr->len; i++)
+    {
+      g_object_unref (g_array_index (arr, GObject *, i));
+    }
+  g_array_free (arr, TRUE);
+}
+
 static void
 salut_muc_manager_init (SalutMucManager *obj)
 {
   SalutMucManagerPrivate *priv = SALUT_MUC_MANAGER_GET_PRIVATE (obj);
   priv->connection = NULL;
   priv->client = NULL;
-  priv->browser = salut_avahi_service_browser_new ("_salut-room._udp");
+  priv->browser = ga_service_browser_new (SALUT_DNSSD_CLIQUE);
 
   /* allocate any data required by the object here */
   priv->text_channels = g_hash_table_new_full(g_direct_hash, g_direct_equal,
                                          NULL, g_object_unref);
 
   priv->room_resolvers = g_hash_table_new_full (g_str_hash, g_str_equal,
-      g_free, g_object_unref);
+      g_free, room_resolver_removed);
+
+#ifdef ENABLE_DBUS_TUBES
   priv->tubes_channels = g_hash_table_new_full (g_direct_hash, g_direct_equal,
       NULL, g_object_unref);
+#endif
+
+  priv->roomlist_channels = NULL;
 }
 
 static void salut_muc_manager_dispose (GObject *object);
@@ -142,7 +163,9 @@ salut_muc_manager_dispose (GObject *object)
 
   tp_channel_factory_iface_close_all (TP_CHANNEL_FACTORY_IFACE (object));
   g_assert (priv->text_channels == NULL);
+#ifdef ENABLE_DBUS_TUBES
   g_assert (priv->tubes_channels == NULL);
+#endif
 
   if (priv->room_resolvers)
     {
@@ -182,6 +205,9 @@ salut_muc_manager_factory_iface_close_all(TpChannelFactoryIface *iface) {
   SalutMucManager *mgr = SALUT_MUC_MANAGER(iface);
   SalutMucManagerPrivate *priv = SALUT_MUC_MANAGER_GET_PRIVATE(mgr);
 
+  g_signal_handlers_disconnect_matched (priv->browser, G_SIGNAL_MATCH_DATA,
+      0, 0, NULL, NULL, mgr);
+
   if (priv->text_channels)
     {
       GHashTable *tmp = priv->text_channels;
@@ -195,12 +221,22 @@ salut_muc_manager_factory_iface_close_all(TpChannelFactoryIface *iface) {
       priv->client = NULL;
     }
 
+#ifdef ENABLE_DBUS_TUBES
   if (priv->tubes_channels != NULL)
     {
       GHashTable *tmp = priv->tubes_channels;
       priv->tubes_channels = NULL;
       g_hash_table_destroy (tmp);
     }
+#endif
+
+  if (priv->roomlist_channels != NULL)
+    {
+      GSList *l = priv->roomlist_channels;
+      priv->roomlist_channels = NULL;
+      g_slist_foreach (l, (GFunc) g_object_unref, NULL);
+      g_slist_free (l);
+    }
 }
 
 static void
@@ -231,6 +267,15 @@ salut_muc_manager_iface_foreach_one(gpointer key,
   f->func(chan, f->data);
 }
 
+static void
+salut_muc_manager_iface_foreach_one_list (TpChannelIface *chan,
+                                          gpointer data)
+{
+  struct foreach_data *f = (struct foreach_data *) data;
+
+  f->func (chan, f->data);
+}
+
 static void
 salut_muc_manager_factory_iface_foreach(TpChannelFactoryIface *iface,
                                         TpChannelFunc func, gpointer data) {
@@ -240,9 +285,15 @@ salut_muc_manager_factory_iface_foreach(TpChannelFactoryIface *iface,
   f.func = func;
   f.data = data;
 
-  g_hash_table_foreach(priv->text_channels, salut_muc_manager_iface_foreach_one, &f);
+  g_hash_table_foreach(priv->text_channels,
+      salut_muc_manager_iface_foreach_one, &f);
+#ifdef ENABLE_DBUS_TUBES
   g_hash_table_foreach (priv->tubes_channels,
       salut_muc_manager_iface_foreach_one, &f);
+#endif
+
+  g_slist_foreach (priv->roomlist_channels,
+      (GFunc) salut_muc_manager_iface_foreach_one_list, &f);
 }
 
 static void
@@ -258,6 +309,7 @@ muc_channel_closed_cb (SalutMucChannel *chan,
       g_object_get(chan, "handle", &handle, NULL);
       DEBUG ("Removing channel with handle %u", handle);
 
+#ifdef ENABLE_DBUS_TUBES
       if (priv->tubes_channels != NULL)
         {
           SalutTubesChannel *tubes;
@@ -267,11 +319,13 @@ muc_channel_closed_cb (SalutMucChannel *chan,
           if (tubes != NULL)
             salut_tubes_channel_close (tubes);
         }
+#endif
 
       g_hash_table_remove (priv->text_channels, GUINT_TO_POINTER (handle));
     }
 }
 
+#ifdef ENABLE_DBUS_TUBES
 /**
  * tubes_channel_closed_cb:
  *
@@ -298,6 +352,7 @@ tubes_channel_closed_cb (SalutTubesChannel *chan, gpointer user_data)
        * but closing the corresponding text channel would be too astonishing */
     }
 }
+#endif
 
 
 static GibberMucConnection *
@@ -312,17 +367,12 @@ _get_connection (SalutMucManager *mgr,
       protocol, parameters, error);
 }
 
-static const gchar **
-_get_connection_parameters (SalutMucManager *mgr,
-                            const gchar *protocol)
-{
-  return gibber_muc_connection_get_required_parameters (protocol);
-}
 
 static SalutMucChannel *
 salut_muc_manager_new_muc_channel (SalutMucManager *mgr,
                                    TpHandle handle,
-                                   GibberMucConnection *connection)
+                                   GibberMucConnection *connection,
+                                   gboolean new_connection)
 {
   SalutMucManagerPrivate *priv = SALUT_MUC_MANAGER_GET_PRIVATE(mgr);
   TpBaseConnection *base_connection = TP_BASE_CONNECTION(priv->connection);
@@ -348,18 +398,21 @@ salut_muc_manager_new_muc_channel (SalutMucManager *mgr,
       "handle", handle,
       "name", name,
       "client", priv->client,
+      "creator", new_connection,
       "xmpp-connection-manager", priv->xmpp_connection_manager,
       NULL);
   g_free (path);
 
-  g_hash_table_insert (priv->text_channels, GUINT_TO_POINTER (handle), chan);
   g_signal_connect (chan, "closed", G_CALLBACK (muc_channel_closed_cb), mgr);
   tp_channel_factory_iface_emit_new_channel (mgr, TP_CHANNEL_IFACE (chan),
       NULL);
 
+  g_hash_table_insert (priv->text_channels, GUINT_TO_POINTER (handle), chan);
+
   return chan;
 }
 
+#ifdef ENABLE_DBUS_TUBES
 /**
  * new_tubes_channel:
  *
@@ -392,8 +445,8 @@ new_tubes_channel (SalutMucManager *self,
       NULL);
 
   g_signal_connect (chan, "closed", (GCallback) tubes_channel_closed_cb, self);
-  tp_channel_factory_iface_emit_new_channel (self, TP_CHANNEL_IFACE (chan),
-      NULL);
+  tp_channel_factory_iface_emit_new_channel (self,
+      TP_CHANNEL_IFACE (chan), NULL);
 
   g_hash_table_insert (priv->tubes_channels, GUINT_TO_POINTER (room), chan);
 
@@ -401,6 +454,7 @@ new_tubes_channel (SalutMucManager *self,
 
   return chan;
 }
+#endif
 
 static gchar *
 _avahi_address_to_string_address (const AvahiAddress *address)
@@ -429,41 +483,52 @@ salut_muc_manager_request_new_muc_channel (SalutMucManager *mgr,
   GError *connection_error = NULL;
   const gchar *room_name;
   GHashTable *params = NULL;
-  SalutAvahiServiceResolver *resolver;
+  GArray *arr;
   gboolean r;
 
   room_name = tp_handle_inspect (room_repo, handle);
-  resolver = g_hash_table_lookup (priv->room_resolvers, room_name);
-  if (resolver != NULL)
+  arr = g_hash_table_lookup (priv->room_resolvers, room_name);
+  if (arr != NULL && arr->len > 0)
     {
       /* This MUC already exists on the network, so we reuse its
        * address */
       AvahiAddress avahi_address;
       uint16_t p;
+      int i;
 
-      if (!salut_avahi_service_resolver_get_address (resolver,
-            &avahi_address, &p))
+      for (i = 0; i < arr->len; i++)
         {
-          DEBUG ("..._get_address failed: creating a new MUC room instead");
-        }
-      else
-        {
-          gchar *address = _avahi_address_to_string_address (&avahi_address);
+           GaServiceResolver *resolver;
+           resolver = g_array_index (arr, GaServiceResolver *, i);
 
-          if (address == NULL)
-            {
-              DEBUG ("stringifying AvahiAddress failed: creating a new MUC "
-                  "room instead");
-            }
-          else
-            {
-              gchar *port = g_strdup_printf ("%u", p);
+           if (!ga_service_resolver_get_address (resolver,
+                 &avahi_address, &p))
+             {
+               DEBUG ("..._get_address failed:"
+                   "creating a new MUC room instead");
+             }
+           else
+             {
+               gchar *address =
+                   _avahi_address_to_string_address (&avahi_address);
 
-              params = g_hash_table_new_full (g_str_hash, g_str_equal, NULL,
-                  g_free);
-              g_hash_table_insert (params, "address", address);
-              g_hash_table_insert (params, "port", port);
-              DEBUG ("found %s port %s for room %s", address, port, room_name);
+              if (address == NULL)
+                {
+                  DEBUG ("stringifying AvahiAddress failed: creating a new MUC "
+                      "room instead");
+                }
+              else
+                {
+                  gchar *port = g_strdup_printf ("%u", p);
+
+                  params = g_hash_table_new_full (g_str_hash, g_str_equal, NULL,
+                      g_free);
+                  g_hash_table_insert (params, "address", address);
+                  g_hash_table_insert (params, "port", port);
+                  DEBUG ("found %s port %s for room %s", address, port,
+                      room_name);
+                  break;
+                }
             }
         }
     }
@@ -499,7 +564,7 @@ salut_muc_manager_request_new_muc_channel (SalutMucManager *mgr,
   DEBUG ("Connect succeeded");
 
   text_chan = salut_muc_manager_new_muc_channel (mgr, handle,
-      connection);
+      connection, params == NULL);
   r = salut_muc_channel_invited (text_chan,
         base_connection->self_handle, NULL, NULL);
   /* Inviting ourselves to a connected channel should always
@@ -509,6 +574,90 @@ salut_muc_manager_request_new_muc_channel (SalutMucManager *mgr,
   return text_chan;
 }
 
+static void
+roomlist_channel_closed_cb (SalutRoomlistChannel *chan,
+                            gpointer data)
+{
+  SalutMucManager *self = SALUT_MUC_MANAGER (data);
+  SalutMucManagerPrivate *priv = SALUT_MUC_MANAGER_GET_PRIVATE (self);
+
+  if (priv->roomlist_channels != NULL)
+    {
+      g_object_unref (chan);
+      priv->roomlist_channels = g_slist_remove (priv->roomlist_channels, chan);
+    }
+}
+
+static void
+add_room_to_roomlist_channel (gpointer key,
+                              gpointer value,
+                              gpointer user_data)
+{
+  const gchar *room_name = (const gchar *) key;
+  SalutRoomlistChannel *roomlist_channel = SALUT_ROOMLIST_CHANNEL (user_data);
+
+  salut_roomlist_channel_add_room (roomlist_channel, room_name);
+}
+
+static SalutRoomlistChannel *
+make_roomlist_channel (SalutMucManager *self,
+                       gpointer request)
+{
+  SalutMucManagerPrivate *priv = SALUT_MUC_MANAGER_GET_PRIVATE (self);
+  TpBaseConnection *conn = (TpBaseConnection *) priv->connection;
+  SalutRoomlistChannel *roomlist_channel;
+  gchar *object_path;
+  static guint cpt = 0;
+
+  object_path = g_strdup_printf ("%s/RoomlistChannel%u",
+      conn->object_path, cpt++);
+
+  roomlist_channel = salut_roomlist_channel_new (priv->connection,
+      object_path);
+
+  g_hash_table_foreach (priv->room_resolvers, add_room_to_roomlist_channel,
+      roomlist_channel);
+
+  priv->roomlist_channels = g_slist_prepend (priv->roomlist_channels,
+      roomlist_channel);
+
+  g_signal_connect (roomlist_channel, "closed",
+      (GCallback) roomlist_channel_closed_cb, self);
+
+  tp_channel_factory_iface_emit_new_channel (self,
+      (TpChannelIface *) roomlist_channel, request);
+
+  g_free (object_path);
+  return roomlist_channel;
+}
+
+static SalutTubesChannel *
+create_tubes_channel (SalutMucManager *self,
+                      TpHandle handle,
+                      GError **error)
+{
+  SalutMucManagerPrivate *priv = SALUT_MUC_MANAGER_GET_PRIVATE (self);
+  SalutMucChannel *text_chan;
+  SalutTubesChannel *tubes_chan;
+
+  text_chan = g_hash_table_lookup (priv->text_channels,
+      GUINT_TO_POINTER (handle));
+
+  if (text_chan == NULL)
+    {
+      DEBUG ("have to create the text channel before the tubes one");
+      text_chan = salut_muc_manager_request_new_muc_channel (self,
+          handle, error);
+      if (text_chan == NULL)
+        return NULL;
+    }
+
+  tubes_chan = new_tubes_channel (self, handle, text_chan);
+  g_assert (tubes_chan != NULL);
+
+  return tubes_chan;
+}
+
 static TpChannelFactoryRequestStatus
 salut_muc_manager_factory_iface_request (TpChannelFactoryIface *iface,
                                          const gchar *chan_type,
@@ -529,7 +678,15 @@ salut_muc_manager_factory_iface_request (TpChannelFactoryIface *iface,
   DEBUG ("MUC request: ctype=%s htype=%u handle=%u", chan_type, handle_type,
       handle);
 
-  /* We only support room handles */
+  if (!tp_strdiff (chan_type, TP_IFACE_CHANNEL_TYPE_ROOM_LIST))
+    {
+      SalutRoomlistChannel *roomlist_channel;
+
+      roomlist_channel = make_roomlist_channel (mgr, request);
+      *ret = TP_CHANNEL_IFACE (roomlist_channel);
+      return TP_CHANNEL_FACTORY_REQUEST_STATUS_CREATED;
+    }
+
   if (handle_type != TP_HANDLE_TYPE_ROOM)
     {
       return TP_CHANNEL_FACTORY_REQUEST_STATUS_NOT_AVAILABLE;
@@ -563,7 +720,8 @@ salut_muc_manager_factory_iface_request (TpChannelFactoryIface *iface,
       g_assert (text_chan != NULL);
       *ret = TP_CHANNEL_IFACE (text_chan);
     }
-  else if (!tp_strdiff (chan_type, SALUT_IFACE_CHANNEL_TYPE_TUBES))
+#ifdef ENABLE_DBUS_TUBES
+  else if (!tp_strdiff (chan_type, TP_IFACE_CHANNEL_TYPE_TUBES))
     {
       SalutTubesChannel *tubes_chan;
 
@@ -573,28 +731,20 @@ salut_muc_manager_factory_iface_request (TpChannelFactoryIface *iface,
       if (tubes_chan != NULL)
         {
           status = TP_CHANNEL_FACTORY_REQUEST_STATUS_EXISTING;
+          *ret = TP_CHANNEL_IFACE (tubes_chan);
         }
       else
         {
-          text_chan = g_hash_table_lookup (priv->text_channels,
-              GUINT_TO_POINTER (handle));
+          tubes_chan = create_tubes_channel (mgr, handle, error);
+          if (tubes_chan == NULL)
+            return TP_CHANNEL_FACTORY_REQUEST_STATUS_ERROR;
 
-          if (text_chan == NULL)
-            {
-              DEBUG ("have to create the text channel before the tubes one");
-              text_chan = salut_muc_manager_request_new_muc_channel (mgr,
-                  handle, error);
-              if (text_chan == NULL)
-                return TP_CHANNEL_FACTORY_REQUEST_STATUS_ERROR;
-            }
+          *ret = TP_CHANNEL_IFACE (tubes_chan);
 
-          tubes_chan = new_tubes_channel (mgr, handle, text_chan);
           status = TP_CHANNEL_FACTORY_REQUEST_STATUS_CREATED;
         }
-
-      g_assert (tubes_chan != NULL);
-      *ret = TP_CHANNEL_IFACE (tubes_chan);
     }
+#endif
   else
     {
       return TP_CHANNEL_FACTORY_REQUEST_STATUS_NOT_IMPLEMENTED;
@@ -628,7 +778,8 @@ invite_stanza_filter (SalutXmppConnectionManager *mgr,
   if (type != GIBBER_STANZA_TYPE_MESSAGE)
     return FALSE;
 
-  return (gibber_xmpp_node_get_child_ns (stanza->node, "x", NS_LLMUC) != NULL);
+  return (gibber_xmpp_node_get_child_ns (stanza->node, "invite",
+        GIBBER_TELEPATHY_NS_CLIQUE) != NULL);
 }
 
 static void
@@ -645,30 +796,23 @@ invite_stanza_callback (SalutXmppConnectionManager *mgr,
       tp_base_connection_get_handles (base_connection, TP_HANDLE_TYPE_ROOM);
   TpHandleRepoIface *contact_repo =
       tp_base_connection_get_handles (base_connection, TP_HANDLE_TYPE_CONTACT);
-  GibberXmppNode *node, *invite, *room_node, *reason_node;
+  GibberXmppNode *invite, *room_node, *reason_node;
   SalutMucChannel *chan;
   const gchar *room = NULL;
   const gchar *reason = NULL;
-  const gchar *protocol = NULL;
   const gchar **params;
   TpHandle room_handle;
-  TpHandle invitor_handle;
+  TpHandle inviter_handle;
   const gchar **p;
   GHashTable *params_hash;
   GibberMucConnection *connection = NULL;
 
-  node = gibber_xmpp_node_get_child_ns (stanza->node, "x", NS_LLMUC);
-  g_assert (node != NULL);
+  invite = gibber_xmpp_node_get_child_ns (stanza->node, "invite",
+      GIBBER_TELEPATHY_NS_CLIQUE);
+  g_assert (invite != NULL);
 
   DEBUG("Got an invitation");
 
-  invite = gibber_xmpp_node_get_child (node, "invite");
-  if (invite == NULL)
-    {
-      DEBUG ("Got invitation, but no invite block!?");
-      return;
-    }
-
   room_node = gibber_xmpp_node_get_child (invite, "roomname");
   if (room_node == NULL)
     {
@@ -684,14 +828,8 @@ invite_stanza_callback (SalutXmppConnectionManager *mgr,
   if (reason == NULL)
     reason = "";
 
-  protocol = gibber_xmpp_node_get_attribute (invite, "protocol");
-  if (protocol == NULL)
-    {
-      DEBUG ("Invalid invitation, (missing protocol) discarding");
-      return;
-    }
-
-  params = _get_connection_parameters (self, protocol);
+  params = gibber_muc_connection_get_required_parameters (
+      GIBBER_TELEPATHY_NS_CLIQUE);
   if (params == NULL)
     {
       DEBUG ("Invalid invitation, (unknown protocol) discarding");
@@ -726,7 +864,8 @@ invite_stanza_callback (SalutXmppConnectionManager *mgr,
 
   if (chan == NULL)
     {
-      connection = _get_connection (self, protocol, params_hash, NULL);
+      connection = _get_connection (self, GIBBER_TELEPATHY_NS_CLIQUE,
+          params_hash, NULL);
       if (connection == NULL)
         {
           DEBUG ("Invalid invitation, (wrong protocol parameters) discarding");
@@ -740,15 +879,22 @@ invite_stanza_callback (SalutXmppConnectionManager *mgr,
           goto discard;
         }
       /* Need to create a new one */
-      chan = salut_muc_manager_new_muc_channel (self, room_handle, connection);
+      chan = salut_muc_manager_new_muc_channel (self, room_handle,
+          connection, FALSE);
     }
 
   /* FIXME handle properly */
   g_assert(chan != NULL);
 
-  invitor_handle = tp_handle_ensure (contact_repo, contact->name, NULL, NULL);
-  salut_muc_channel_invited (chan, invitor_handle, reason, NULL);
-  tp_handle_unref (contact_repo, invitor_handle);
+  inviter_handle = tp_handle_ensure (contact_repo, contact->name, NULL, NULL);
+
+#ifdef ENABLE_OLPC
+  salut_connection_olpc_observe_invitation (priv->connection, room_handle,
+      inviter_handle, invite);
+#endif
+
+  salut_muc_channel_invited (chan, inviter_handle, reason, NULL);
+  tp_handle_unref (contact_repo, inviter_handle);
 
 discard:
   if (params_hash != NULL)
@@ -781,29 +927,33 @@ salut_muc_manager_new (SalutConnection *connection,
 }
 
 static void
-browser_found (SalutAvahiServiceBrowser *browser,
+add_room_foreach (SalutRoomlistChannel *roomlist_channel,
+                  const gchar *room)
+{
+  salut_roomlist_channel_add_room (roomlist_channel, room);
+}
+
+static void
+browser_found (GaServiceBrowser *browser,
                AvahiIfIndex interface,
                AvahiProtocol protocol,
                const char *name,
                const char *type,
                const char *domain,
-               SalutAvahiLookupResultFlags flags,
+               GaLookupResultFlags flags,
                gpointer userdata)
 {
   SalutMucManager *self = SALUT_MUC_MANAGER (userdata);
   SalutMucManagerPrivate *priv = SALUT_MUC_MANAGER_GET_PRIVATE (self);
-  SalutAvahiServiceResolver *resolver;
+  GArray *arr;
+  GaServiceResolver *resolver;
   GError *error = NULL;
 
-  resolver = g_hash_table_lookup (priv->room_resolvers, name);
-  if (resolver != NULL)
-    return;
-
   DEBUG ("found room: %s.%s.%s", name, type, domain);
-  resolver = salut_avahi_service_resolver_new (interface, protocol,
+  resolver = ga_service_resolver_new (interface, protocol,
       name, type, domain, protocol, 0);
 
-  if (!salut_avahi_service_resolver_attach (resolver, priv->client,
+  if (!ga_service_resolver_attach (resolver, priv->client,
         &error))
     {
       DEBUG ("resolver attach failed: %s", error->message);
@@ -812,28 +962,112 @@ browser_found (SalutAvahiServiceBrowser *browser,
       return;
     }
 
-  g_hash_table_insert (priv->room_resolvers, g_strdup (name), resolver);
+  arr = g_hash_table_lookup (priv->room_resolvers, name);
+  if (arr == NULL)
+    {
+      arr = g_array_new (FALSE, FALSE, sizeof (GObject *));
+      g_hash_table_insert (priv->room_resolvers, g_strdup (name), arr);
+      g_slist_foreach (priv->roomlist_channels, (GFunc) add_room_foreach,
+          (gchar *) name);
+    }
+  g_array_append_val (arr, resolver);
 }
 
 static void
-browser_removed (SalutAvahiServiceBrowser *browser,
+remove_room_foreach (SalutRoomlistChannel *roomlist_channel,
+                     const gchar *room)
+{
+  salut_roomlist_channel_remove_room (roomlist_channel, room);
+}
+
+static void
+browser_removed (GaServiceBrowser *browser,
                  AvahiIfIndex interface,
                  AvahiProtocol protocol,
                  const char *name,
                  const char *type,
                  const char *domain,
-                 SalutAvahiLookupResultFlags flags,
+                 GaLookupResultFlags flags,
                  gpointer userdata)
 {
   SalutMucManager *self = SALUT_MUC_MANAGER (userdata);
   SalutMucManagerPrivate *priv = SALUT_MUC_MANAGER_GET_PRIVATE (self);
+  TpBaseConnection *base_connection = TP_BASE_CONNECTION (priv->connection);
+  TpHandleRepoIface *room_repo =
+      tp_base_connection_get_handles (base_connection, TP_HANDLE_TYPE_ROOM);
+  TpHandle handle;
+  SalutMucChannel *muc;
+  GArray *arr;
+  int i;
+
+
+  arr = g_hash_table_lookup (priv->room_resolvers, name);
+
+  if (arr == NULL) {
+    DEBUG ("Browser removed for %s, but didn't have any resolvers", name);
+    return;
+  }
+
+  for (i = 0; i < arr->len; i++)
+    {
+      GaServiceResolver *resolver;
+      AvahiIfIndex r_interface;
+      AvahiProtocol r_protocol;
+      gchar *r_name;
+      gchar *r_type;
+      gchar *r_domain;
+
+      resolver = g_array_index (arr, GaServiceResolver *, i);
+      g_object_get ((gpointer) resolver,
+          "interface", &r_interface,
+          "protocol", &r_protocol,
+          "name", &r_name,
+          "type", &r_type,
+          "domain", &r_domain,
+          NULL);
+      if (interface == r_interface
+          && protocol == r_protocol
+          && !tp_strdiff (name, r_name)
+          && !tp_strdiff (type, r_type)
+          && !tp_strdiff (domain, r_domain))
+        {
+          g_free (r_name);
+          g_free (r_type);
+          g_free (r_domain);
+          g_object_unref (resolver);
+          g_array_remove_index_fast (arr, i);
+          break;
+        }
+
+      g_free (r_name);
+      g_free (r_type);
+      g_free (r_domain);
+    }
+
+  if (arr->len > 0)
+    return;
 
   DEBUG ("remove room: %s.%s.%s", name, type, domain);
+  g_slist_foreach (priv->roomlist_channels, (GFunc) remove_room_foreach,
+      (gchar *) name);
+
   g_hash_table_remove (priv->room_resolvers, name);
+
+  /* Do we have to re-announce this room ? */
+  handle = tp_handle_lookup (room_repo, name, NULL, NULL);
+  if (handle == 0)
+    return;
+
+  muc = g_hash_table_lookup (priv->text_channels, GUINT_TO_POINTER (handle));
+  if (muc == NULL)
+    return;
+
+  DEBUG ("We know this room %s. Try to re-announce it", name);
+  salut_muc_channel_publish_service (muc);
 }
 
 static void
-browser_failed (SalutAvahiServiceBrowser *browser,
+browser_failed (GaServiceBrowser *browser,
                 GError *error,
                 gpointer userdata)
 {
@@ -843,7 +1077,7 @@ browser_failed (SalutAvahiServiceBrowser *browser,
 
 gboolean
 salut_muc_manager_start (SalutMucManager *self,
-                         SalutAvahiClient *client,
+                         GaClient *client,
                          GError **error)
 {
   SalutMucManagerPrivate *priv = SALUT_MUC_MANAGER_GET_PRIVATE (self);
@@ -861,7 +1095,7 @@ salut_muc_manager_start (SalutMucManager *self,
   g_signal_connect (priv->browser, "failure",
                    G_CALLBACK (browser_failed), self);
 
-  if (!salut_avahi_service_browser_attach (priv->browser, priv->client,
+  if (!ga_service_browser_attach (priv->browser, priv->client,
         error))
     {
       DEBUG ("browser attach failed");
@@ -870,3 +1104,71 @@ salut_muc_manager_start (SalutMucManager *self,
 
   return TRUE;
 }
+
+SalutMucChannel *
+salut_muc_manager_get_text_channel (SalutMucManager *self,
+                                    TpHandle handle)
+{
+  SalutMucManagerPrivate *priv = SALUT_MUC_MANAGER_GET_PRIVATE (self);
+  SalutMucChannel *muc;
+
+  muc = g_hash_table_lookup (priv->text_channels, GUINT_TO_POINTER (handle));
+  if (muc == NULL)
+    return NULL;
+
+  g_object_ref (muc);
+  return muc;
+}
+
+void
+salut_muc_manager_handle_si_stream_request (SalutMucManager *self,
+                                            GibberBytestreamIface *bytestream,
+                                            TpHandle room_handle,
+                                            const gchar *stream_id,
+                                            GibberXmppStanza *msg)
+{
+  SalutMucManagerPrivate *priv = SALUT_MUC_MANAGER_GET_PRIVATE (self);
+  TpHandleRepoIface *room_repo = tp_base_connection_get_handles (
+     (TpBaseConnection*) priv->connection, TP_HANDLE_TYPE_ROOM);
+  SalutTubesChannel *chan = NULL;
+
+  g_return_if_fail (tp_handle_is_valid (room_repo, room_handle, NULL));
+
+#ifdef ENABLE_DBUS_TUBES
+  chan = g_hash_table_lookup (priv->tubes_channels,
+      GUINT_TO_POINTER (room_handle));
+#endif
+  if (chan == NULL)
+    {
+      GError e = { GIBBER_XMPP_ERROR, XMPP_ERROR_BAD_REQUEST,
+          "No tubes channel available for this MUC" };
+
+      DEBUG ("tubes channel doesn't exist for muc %d", room_handle);
+      gibber_bytestream_iface_close (bytestream, &e);
+      return;
+    }
+
+  salut_tubes_channel_bytestream_offered (chan, bytestream, msg);
+}
+
+SalutTubesChannel *
+salut_muc_manager_ensure_tubes_channel (SalutMucManager *self,
+                                        TpHandle handle)
+{
+  SalutMucManagerPrivate *priv = SALUT_MUC_MANAGER_GET_PRIVATE (self);
+  SalutTubesChannel *tubes_chan;
+
+  tubes_chan = g_hash_table_lookup (priv->tubes_channels,
+      GUINT_TO_POINTER (handle));
+  if (tubes_chan != NULL)
+    {
+      g_object_ref (tubes_chan);
+      return tubes_chan;
+    }
+
+  tubes_chan = create_tubes_channel (self, handle, NULL);
+  g_assert (tubes_chan != NULL);
+  g_object_ref (tubes_chan);
+
+  return tubes_chan;
+}
