@@ -23,10 +23,13 @@
 #include "config.h"
 
 #include <glib-object.h>
+#include <avahi-gobject/ga-client.h>
 
 #include <telepathy-glib/handle-repo.h>
 
-#include "salut-avahi-client.h"
+#include <gibber/gibber-xmpp-stanza.h>
+
+#include "salut-connection.h"
 #include "salut-presence.h"
 
 G_BEGIN_DECLS
@@ -71,7 +74,7 @@ GType salut_self_get_type(void);
 #define SALUT_SELF_GET_CLASS(obj) \
   (G_TYPE_INSTANCE_GET_CLASS ((obj), SALUT_TYPE_SELF, SalutSelfClass))
 
-SalutSelf *salut_self_new (SalutAvahiClient *client,
+SalutSelf *salut_self_new (SalutConnection *conn, GaClient *client,
     TpHandleRepoIface *room_repo, const gchar *nickname,
     const gchar *first_name, const gchar *last_name, const gchar *jid,
     const gchar *email, const gchar *published_name,
@@ -79,8 +82,7 @@ SalutSelf *salut_self_new (SalutAvahiClient *client,
     const gchar *olpc_color);
 
 /* Start announcing our presence on the network */
-gboolean
-salut_self_announce (SalutSelf *self, gint port, GError **error);
+gboolean salut_self_announce (SalutSelf *self, gint port, GError **error);
 
 gboolean salut_self_set_presence (SalutSelf *self,
     SalutPresenceId status, const gchar *message, GError **error);
@@ -99,12 +101,18 @@ gboolean salut_self_set_olpc_properties (SalutSelf *self,
 
 gboolean salut_self_merge_olpc_activity_properties (SalutSelf *self,
     TpHandle handle,
-    const gchar **color, const gchar **name, const gchar **type);
+    const gchar **color, const gchar **name, const gchar **type,
+    const gchar **tags, gboolean *is_private);
 
 gboolean salut_self_set_olpc_activity_properties (SalutSelf *self,
     TpHandle handle,
     const gchar *color, const gchar *name, const gchar *type,
-    GError **error);
+    const gchar *tags, gboolean is_private, GError **error);
+
+gboolean salut_self_olpc_activity_properties_updated (SalutSelf *self,
+    TpHandle handle,
+    const gchar *color, const gchar *name, const gchar *type,
+    const gchar *tags, gboolean is_private);
 
 gboolean salut_self_set_olpc_activities (SalutSelf *self,
     GHashTable *act_id_to_room, GError **error);
@@ -117,6 +125,9 @@ typedef void (*SalutSelfOLPCActivityFunc)
 
 void salut_self_foreach_olpc_activity (SalutSelf *self,
     SalutSelfOLPCActivityFunc foreach, gpointer user_data);
+
+void salut_self_olpc_augment_invitation (SalutSelf *self,
+    TpHandle room, TpHandle contact, GibberXmppNode *invite_node);
 #endif
 
 G_END_DECLS
