@@ -43,3 +43,15 @@ sha1_hex (const guint8 *bytes, guint len)
 
   return hex_hash;
 }
+
+void
+sha1_bin (const gchar *bytes,
+          guint len,
+          guchar out[SHA1_HASH_SIZE])
+{
+  SHA1Context sc;
+
+  SHA1Init (&sc);
+  SHA1Update (&sc, bytes, len);
+  SHA1Final (&sc, (uint8_t *) out);
+}
