@@ -22,26 +22,25 @@
 #include <stdlib.h>
 #include <string.h>
 #include <time.h>
+#include <errno.h>
 
 #include <glib.h>
+#include <glib/gstdio.h>
 
 #include <dbus/dbus-glib.h>
 #include <dbus/dbus-glib-lowlevel.h>
 #include <telepathy-glib/util.h>
 
 #include <gibber/gibber-bytestream-ibb.h>
+#include <gibber/gibber-bytestream-muc.h>
 #include <gibber/gibber-muc-connection.h>
 
 #define DEBUG_FLAG DEBUG_TUBES
 #include "debug.h"
-#include "extensions/extensions.h"
 #include "salut-connection.h"
-#include "namespaces.h"
 #include "tube-iface.h"
+#include "sha1/sha1-util.h"
 
-/*
-#include "bytestream-factory.h"
-*/
 
 static void
 tube_iface_init (gpointer g_iface, gpointer iface_data);
@@ -90,7 +89,7 @@ struct _SalutTubeDBusPrivate
   TpHandle self_handle;
   GibberMucConnection *muc_connection;
   guint id;
-  GibberBytestreamIBB *bytestream;
+  GibberBytestreamIface *bytestream;
   gchar *stream_id;
   TpHandle initiator;
   gchar *service;
@@ -100,21 +99,28 @@ struct _SalutTubeDBusPrivate
   gchar *dbus_local_name;
   /* the address that we are listening for D-Bus connections on */
   gchar *dbus_srv_addr;
+  /* the path of the UNIX socket used by the D-Bus server */
+  gchar *socket_path;
   /* the server that's listening on dbus_srv_addr */
   DBusServer *dbus_srv;
   /* the connection to dbus_srv from a local client, or NULL */
   DBusConnection *dbus_conn;
-  /* mapping of contact handle -> D-Bus name */
+  /* mapping of contact handle -> D-Bus name (NULL for 1-1 D-Bus tubes) */
   GHashTable *dbus_names;
 
+  /* Message reassembly buffer (CONTACT tubes only) */
+  GString *reassembly_buffer;
+  /* Number of bytes that will be in the next message, 0 if unknown */
+  guint32 reassembly_bytes_needed;
+
   gboolean dispose_has_run;
 };
 
 #define SALUT_TUBE_DBUS_GET_PRIVATE(obj) \
     ((SalutTubeDBusPrivate *) obj->priv)
 
-static void data_received_cb (GibberBytestreamIBB *ibb, const gchar *from,
-    GString *data, gpointer user_data);
+static void data_received_cb (GibberBytestreamIface *bytestream,
+    const gchar *from, GString *data, gpointer user_data);
 
 /*
  * Characters used are permissible both in filenames and in D-Bus names. (See
@@ -135,6 +141,57 @@ generate_ascii_string (guint len,
     buf[i] = chars[g_random_int_range (0, 64)];
 }
 
+static gchar *
+generate_dbus_unique_name (const gchar *nick)
+{
+  gchar *encoded, *result;
+  size_t len;
+  guint i;
+
+  len = strlen (nick);
+
+  if (len <= 186)
+    {
+      encoded = g_base64_encode ((const guchar *) nick, strlen (nick));
+    }
+  else
+    {
+      guchar sha1[20];
+      GString *tmp;
+
+      sha1_bin (nick, len, sha1);
+      tmp = g_string_sized_new (169 + 20);
+
+      g_string_append_len (tmp, nick, 169);
+      g_string_append_len (tmp, (const gchar *) sha1, 20);
+
+      encoded = g_base64_encode ((const guchar *) tmp->str, tmp->len);
+
+      g_string_free (tmp, TRUE);
+    }
+
+  for (i = 0; encoded[i] != '\0'; i++)
+    {
+      switch (encoded[i])
+        {
+          case '+':
+            encoded[i] = '_';
+            break;
+          case '/':
+            encoded[i] = '-';
+            break;
+          case '=':
+            encoded[i] = 'A';
+            break;
+        }
+    }
+
+  result = g_strdup_printf (":2.%s", encoded);
+
+  g_free (encoded);
+  return result;
+}
+
 struct _find_contact_data
 {
   const gchar *contact;
@@ -150,6 +207,8 @@ filter_cb (DBusConnection *conn,
   SalutTubeDBusPrivate *priv = SALUT_TUBE_DBUS_GET_PRIVATE (tube);
   gchar *marshalled = NULL;
   gint len;
+  const gchar *destination;
+  dbus_uint32_t serial;
 
   if (dbus_message_get_type (msg) == DBUS_MESSAGE_TYPE_SIGNAL &&
       !tp_strdiff (dbus_message_get_interface (msg),
@@ -164,12 +223,27 @@ filter_cb (DBusConnection *conn,
       goto out;
     }
 
-  dbus_message_set_sender (msg, priv->dbus_local_name);
+  if (priv->dbus_local_name != NULL)
+    {
+      dbus_message_set_sender (msg, priv->dbus_local_name);
+    }
 
   if (!dbus_message_marshal (msg, &marshalled, &len))
     goto out;
 
-  gibber_bytestream_ibb_send (priv->bytestream, len, marshalled);
+  gibber_bytestream_iface_send (priv->bytestream, len, marshalled);
+
+  if (GIBBER_IS_BYTESTREAM_MUC (priv->bytestream))
+    {
+      /* In a Salut MUC we never receive messages we sent, so we need to
+       * artificially receive our own messages. */
+      destination = dbus_message_get_destination (msg);
+      if (destination == NULL || !tp_strdiff (priv->dbus_local_name,
+            destination))
+        {
+          dbus_connection_send (priv->dbus_conn, msg, &serial);
+        }
+    }
 
 out:
   if (marshalled != NULL)
@@ -202,23 +276,46 @@ new_connection_cb (DBusServer *server,
 static void
 tube_dbus_open (SalutTubeDBus *self)
 {
+#define SERVER_LISTEN_MAX_TRIES 5
   SalutTubeDBusPrivate *priv = SALUT_TUBE_DBUS_GET_PRIVATE (self);
-  DBusError error = {0,};
-  gchar suffix[8];
+  guint i;
 
   g_signal_connect (priv->bytestream, "data-received",
       G_CALLBACK (data_received_cb), self);
 
-  generate_ascii_string (8, suffix);
-  priv->dbus_srv_addr = g_strdup_printf (
-      "unix:path=/tmp/dbus-salut-%.8s", suffix);
-  DEBUG ("listening on %s", priv->dbus_srv_addr);
-  priv->dbus_srv = dbus_server_listen (priv->dbus_srv_addr, &error);
+  for (i = 0; i < SERVER_LISTEN_MAX_TRIES; i++)
+    {
+      gchar suffix[8];
+      DBusError error;
 
-  /* XXX: if dbus_server_listen fails, we should retry with different
-   * addresses, then close the tube if we give up
-   */
-  g_assert (priv->dbus_srv);
+      g_free (priv->dbus_srv_addr);
+      g_free (priv->socket_path);
+
+      generate_ascii_string (8, suffix);
+      priv->socket_path = g_strdup_printf ("%s/dbus-salut-%.8s",
+          g_get_tmp_dir (), suffix);
+      priv->dbus_srv_addr = g_strdup_printf ("unix:path=%s",
+          priv->socket_path);
+
+      dbus_error_init (&error);
+      priv->dbus_srv = dbus_server_listen (priv->dbus_srv_addr, &error);
+
+      if (priv->dbus_srv_addr != NULL)
+        break;
+
+      DEBUG ("dbus_server_listen failed (try %u): %s: %s", i, error.name,
+          error.message);
+      dbus_error_free (&error);
+    }
+
+  if (priv->dbus_srv_addr == NULL)
+    {
+      DEBUG ("all attempts failed. Close the tube");
+      salut_tube_iface_accept (SALUT_TUBE_IFACE (self), NULL);
+      return;
+    }
+
+  DEBUG ("listening on %s", priv->dbus_srv_addr);
 
   dbus_server_set_new_connection_function (priv->dbus_srv, new_connection_cb,
       self, NULL);
@@ -230,23 +327,8 @@ salut_tube_dbus_init (SalutTubeDBus *self)
 {
   SalutTubeDBusPrivate *priv = G_TYPE_INSTANCE_GET_PRIVATE (self,
       SALUT_TYPE_TUBE_DBUS, SalutTubeDBusPrivate);
-  gchar suffix[8];
 
   self->priv = priv;
-
-  priv->bytestream = NULL;
-  priv->muc_connection = NULL;
-  priv->dispose_has_run = FALSE;
-
-  /* XXX: check this doesn't clash with other bus names */
-  /* this has to contain at least two dot-separated components */
-
-  generate_ascii_string (8, suffix);
-  priv->dbus_local_name = g_strdup_printf (":1.%.8s", suffix);
-  priv->dbus_names = g_hash_table_new_full (g_direct_hash, g_direct_equal,
-      NULL, g_free);
-
-  DEBUG ("local name: %s", priv->dbus_local_name);
 }
 
 static void
@@ -260,42 +342,44 @@ unref_handle_foreach (gpointer key,
   tp_handle_unref (contact_repo, handle);
 }
 
-static SalutTubeState
+static TpTubeState
 get_tube_state (SalutTubeDBus *self)
 {
   SalutTubeDBusPrivate *priv = SALUT_TUBE_DBUS_GET_PRIVATE (self);
-  GibberBytestreamIBBState bytestream_state;
+  GibberBytestreamState bytestream_state;
 
   if (priv->bytestream == NULL)
     /* bytestream not yet created as we're waiting for the SI reply */
-    return SALUT_TUBE_STATE_REMOTE_PENDING;
+    return TP_TUBE_STATE_REMOTE_PENDING;
 
   g_object_get (priv->bytestream, "state", &bytestream_state, NULL);
 
-  if (bytestream_state == GIBBER_BYTESTREAM_IBB_STATE_OPEN)
-    return SALUT_TUBE_STATE_OPEN;
-
-  else if (bytestream_state == GIBBER_BYTESTREAM_IBB_STATE_LOCAL_PENDING ||
-      bytestream_state == GIBBER_BYTESTREAM_IBB_STATE_ACCEPTED)
-    return SALUT_TUBE_STATE_LOCAL_PENDING;
-
-  else if (bytestream_state == GIBBER_BYTESTREAM_IBB_STATE_INITIATING)
-    return SALUT_TUBE_STATE_REMOTE_PENDING;
-
-  else
-    g_assert_not_reached ();
-  return SALUT_TUBE_STATE_REMOTE_PENDING;
+  switch (bytestream_state)
+    {
+      case GIBBER_BYTESTREAM_STATE_OPEN:
+        return TP_TUBE_STATE_OPEN;
+        break;
+      case GIBBER_BYTESTREAM_STATE_LOCAL_PENDING:
+      case GIBBER_BYTESTREAM_STATE_ACCEPTED:
+        return TP_TUBE_STATE_LOCAL_PENDING;
+        break;
+      case GIBBER_BYTESTREAM_STATE_INITIATING:
+        return TP_TUBE_STATE_REMOTE_PENDING;
+        break;
+      default:
+        g_assert_not_reached ();
+    }
 }
 
 static void
-bytestream_state_changed_cb (GibberBytestreamIBB *bytestream,
-                             GibberBytestreamIBBState state,
+bytestream_state_changed_cb (GibberBytestreamIface *bytestream,
+                             GibberBytestreamState state,
                              gpointer user_data)
 {
   SalutTubeDBus *self = SALUT_TUBE_DBUS (user_data);
   SalutTubeDBusPrivate *priv = SALUT_TUBE_DBUS_GET_PRIVATE (self);
 
-  if (state == GIBBER_BYTESTREAM_IBB_STATE_CLOSED)
+  if (state == GIBBER_BYTESTREAM_STATE_CLOSED)
     {
       if (priv->bytestream != NULL)
         {
@@ -305,7 +389,7 @@ bytestream_state_changed_cb (GibberBytestreamIBB *bytestream,
 
       g_signal_emit (G_OBJECT (self), signals[CLOSED], 0);
     }
-  else if (state == GIBBER_BYTESTREAM_IBB_STATE_OPEN)
+  else if (state == GIBBER_BYTESTREAM_STATE_OPEN)
     {
       tube_dbus_open (self);
       g_signal_emit (G_OBJECT (self), signals[OPENED], 0);
@@ -325,7 +409,7 @@ salut_tube_dbus_dispose (GObject *object)
 
   if (priv->bytestream)
     {
-      gibber_bytestream_ibb_close (priv->bytestream);
+      gibber_bytestream_iface_close (priv->bytestream, NULL);
     }
 
   if (priv->dbus_conn)
@@ -337,11 +421,21 @@ salut_tube_dbus_dispose (GObject *object)
   if (priv->dbus_srv)
     dbus_server_unref (priv->dbus_srv);
 
-  if (priv->dbus_srv_addr)
-    g_free (priv->dbus_srv_addr);
+  if (priv->socket_path != NULL)
+    {
+      if (g_unlink (priv->socket_path) != 0)
+        {
+          DEBUG ("unlink of %s failed: %s", priv->socket_path,
+              g_strerror (errno));
+        }
+    }
 
-  if (priv->dbus_local_name)
-    g_free (priv->dbus_local_name);
+  g_free (priv->dbus_srv_addr);
+  priv->dbus_srv_addr = NULL;
+  g_free (priv->socket_path);
+  priv->socket_path = NULL;
+  g_free (priv->dbus_local_name);
+  priv->dbus_local_name = NULL;
 
   if (priv->dbus_names)
     {
@@ -358,6 +452,9 @@ salut_tube_dbus_dispose (GObject *object)
       priv->muc_connection = NULL;
     }
 
+  if (priv->reassembly_buffer)
+    g_string_free (priv->reassembly_buffer, TRUE);
+
   tp_handle_unref (contact_repo, priv->initiator);
 
   priv->dispose_has_run = TRUE;
@@ -415,7 +512,7 @@ salut_tube_dbus_get_property (GObject *object,
         g_value_set_string (value, priv->stream_id);
         break;
       case PROP_TYPE:
-        g_value_set_uint (value, SALUT_TUBE_TYPE_DBUS);
+        g_value_set_uint (value, TP_TUBE_TYPE_DBUS);
         break;
       case PROP_INITIATOR:
         g_value_set_uint (value, priv->initiator);
@@ -477,13 +574,13 @@ salut_tube_dbus_set_property (GObject *object,
       case PROP_BYTESTREAM:
         if (priv->bytestream == NULL)
           {
-            GibberBytestreamIBBState state;
+            GibberBytestreamState state;
 
             priv->bytestream = g_value_get_object (value);
             g_object_ref (priv->bytestream);
 
             g_object_get (priv->bytestream, "state", &state, NULL);
-            if (state == GIBBER_BYTESTREAM_IBB_STATE_OPEN)
+            if (state == GIBBER_BYTESTREAM_STATE_OPEN)
               {
                 tube_dbus_open (self);
               }
@@ -492,10 +589,6 @@ salut_tube_dbus_set_property (GObject *object,
                 G_CALLBACK (bytestream_state_changed_cb), self);
           }
         break;
-      case PROP_STREAM_ID:
-        g_free (priv->stream_id);
-        priv->stream_id = g_value_dup_string (value);
-        break;
       case PROP_INITIATOR:
         priv->initiator = g_value_get_uint (value);
         break;
@@ -543,46 +636,61 @@ salut_tube_dbus_constructor (GType type,
   if (priv->handle_type == TP_HANDLE_TYPE_ROOM)
     {
       /*
-       * We have to create an IBB bytestream that will be
+       * We have to create an MUC bytestream that will be
        * used by this MUC tube to communicate.
        *
        * We don't create the bytestream of private D-Bus tube yet.
        * It will be when we'll receive the answer of the SI request
        */
-      GibberBytestreamIBB *bytestream;
-      GibberBytestreamIBBState state;
+      GibberBytestreamMuc *bytestream;
+      GibberBytestreamState state;
       const gchar *peer_id;
 
       g_assert (priv->muc_connection != NULL);
-      g_assert (priv->stream_id != NULL);
+
+      priv->dbus_names = g_hash_table_new_full (g_direct_hash, g_direct_equal,
+          NULL, g_free);
+
+      priv->dbus_local_name = generate_dbus_unique_name (priv->conn->name);
+
+      DEBUG ("local name: %s", priv->dbus_local_name);
 
       peer_id = tp_handle_inspect (handles_repo, priv->handle);
       if (priv->initiator == priv->self_handle)
         {
           /* We create this tube, bytestream is open */
-          state = GIBBER_BYTESTREAM_IBB_STATE_OPEN;
+          state = GIBBER_BYTESTREAM_STATE_OPEN;
         }
       else
         {
           /* We don't create this tube, bytestream is local pending */
-          state = GIBBER_BYTESTREAM_IBB_STATE_LOCAL_PENDING;
+          state = GIBBER_BYTESTREAM_STATE_LOCAL_PENDING;
         }
 
-      bytestream = g_object_new (GIBBER_TYPE_BYTESTREAM_IBB,
+      bytestream = g_object_new (GIBBER_TYPE_BYTESTREAM_MUC,
             "muc-connection", priv->muc_connection,
-            "stream-id", priv->stream_id,
             "state", state,
             "self-id", priv->conn->name,
             "peer-id", peer_id,
-            "stream-init-id", NULL,
             NULL);
 
+      g_object_get (bytestream, "stream-id", &priv->stream_id, NULL);
+
       g_object_set (self, "bytestream", bytestream, NULL);
+      g_object_unref (bytestream);
     }
   else
     {
       /* Private tube */
       g_assert (priv->muc_connection == NULL);
+
+      /* The D-Bus names mapping is used in muc tubes only */
+      priv->dbus_local_name = NULL;
+      priv->dbus_names = NULL;
+
+      /* For contact tubes we need to be able to reassemble messages. */
+      priv->reassembly_buffer = g_string_new ("");
+      priv->reassembly_bytes_needed = 0;
     }
 
   return obj;
@@ -604,57 +712,26 @@ salut_tube_dbus_class_init (SalutTubeDBusClass *salut_tube_dbus_class)
   object_class->dispose = salut_tube_dbus_dispose;
   object_class->finalize = salut_tube_dbus_finalize;
 
-  param_spec = g_param_spec_object (
-      "connection",
-      "SalutConnection object",
-      "Salut connection object that owns this D-Bus tube object.",
-      SALUT_TYPE_CONNECTION,
-      G_PARAM_CONSTRUCT_ONLY |
-      G_PARAM_READWRITE |
-      G_PARAM_STATIC_NAME |
-      G_PARAM_STATIC_NICK |
-      G_PARAM_STATIC_BLURB);
-  g_object_class_install_property (object_class, PROP_CONNECTION, param_spec);
-
- param_spec = g_param_spec_uint (
-      "handle",
-      "Handle",
-      "The TpHandle associated with the tubes channel that"
-      "owns this D-Bus tube object.",
-      0, G_MAXUINT32, 0,
-      G_PARAM_CONSTRUCT_ONLY |
-      G_PARAM_READWRITE |
-      G_PARAM_STATIC_NAME |
-      G_PARAM_STATIC_NICK |
-      G_PARAM_STATIC_BLURB);
-  g_object_class_install_property (object_class, PROP_HANDLE, param_spec);
-
-  param_spec = g_param_spec_uint (
-      "handle-type",
-      "Handle type",
-      "The TpHandleType of the handle associated with the tubes channel that"
-      "owns this D-Bus tube object.",
-      0, G_MAXUINT32, 0,
-      G_PARAM_CONSTRUCT_ONLY |
-      G_PARAM_READWRITE |
-      G_PARAM_STATIC_NAME |
-      G_PARAM_STATIC_NICK |
-      G_PARAM_STATIC_BLURB);
-  g_object_class_install_property (object_class, PROP_HANDLE_TYPE,
-      param_spec);
-
-  param_spec = g_param_spec_uint (
-      "self-handle",
-      "Self handle",
-      "The handle to use for ourself. This can be different from the "
-      "connection's self handle if our handle is a room handle.",
-      0, G_MAXUINT, 0,
-      G_PARAM_CONSTRUCT_ONLY |
-      G_PARAM_READWRITE |
-      G_PARAM_STATIC_NAME |
-      G_PARAM_STATIC_NICK |
-      G_PARAM_STATIC_BLURB);
-  g_object_class_install_property (object_class, PROP_SELF_HANDLE, param_spec);
+  g_object_class_override_property (object_class, PROP_CONNECTION,
+    "connection");
+  g_object_class_override_property (object_class, PROP_HANDLE,
+    "handle");
+  g_object_class_override_property (object_class, PROP_HANDLE_TYPE,
+    "handle-type");
+  g_object_class_override_property (object_class, PROP_SELF_HANDLE,
+    "self-handle");
+  g_object_class_override_property (object_class, PROP_ID,
+    "id");
+  g_object_class_override_property (object_class, PROP_TYPE,
+    "type");
+  g_object_class_override_property (object_class, PROP_INITIATOR,
+    "initiator");
+  g_object_class_override_property (object_class, PROP_SERVICE,
+    "service");
+  g_object_class_override_property (object_class, PROP_PARAMETERS,
+    "parameters");
+  g_object_class_override_property (object_class, PROP_STATE,
+    "state");
 
   param_spec = g_param_spec_object (
       "muc-connection",
@@ -670,24 +747,11 @@ salut_tube_dbus_class_init (SalutTubeDBusClass *salut_tube_dbus_class)
   g_object_class_install_property (object_class, PROP_MUC_CONNECTION,
       param_spec);
 
-  param_spec = g_param_spec_uint (
-      "id",
-      "id",
-      "The unique identifier of this tube",
-      0, G_MAXUINT32, 0,
-      G_PARAM_CONSTRUCT_ONLY |
-      G_PARAM_READWRITE |
-      G_PARAM_STATIC_NAME |
-      G_PARAM_STATIC_NICK |
-      G_PARAM_STATIC_BLURB);
-  g_object_class_install_property (object_class, PROP_ID, param_spec);
-
   param_spec = g_param_spec_object (
       "bytestream",
-      "SalutBytestreamIBB object",
-      "Salut bytestream IBB object used for streaming data for this D-Bus"
-      "tube object.",
-      GIBBER_TYPE_BYTESTREAM_IBB,
+      "Object implementing the GibberBytestreamIface interface",
+      "Bytestream object used for streaming data for this D-Bus",
+      G_TYPE_OBJECT,
       G_PARAM_READWRITE |
       G_PARAM_STATIC_NAME |
       G_PARAM_STATIC_NICK |
@@ -699,71 +763,12 @@ salut_tube_dbus_class_init (SalutTubeDBusClass *salut_tube_dbus_class)
       "stream id",
       "The identifier of this tube's bytestream",
       "",
-      G_PARAM_CONSTRUCT_ONLY |
-      G_PARAM_READWRITE |
+      G_PARAM_READABLE |
       G_PARAM_STATIC_NAME |
       G_PARAM_STATIC_NICK |
       G_PARAM_STATIC_BLURB);
   g_object_class_install_property (object_class, PROP_STREAM_ID, param_spec);
 
-  param_spec = g_param_spec_uint (
-      "type",
-      "Tube type",
-      "The TpTubeType this D-Bus tube object.",
-      0, G_MAXUINT32, SALUT_TUBE_TYPE_DBUS,
-      G_PARAM_READABLE |
-      G_PARAM_STATIC_NAME |
-      G_PARAM_STATIC_NICK |
-      G_PARAM_STATIC_BLURB);
-  g_object_class_install_property (object_class, PROP_TYPE, param_spec);
-
-  param_spec = g_param_spec_uint (
-      "initiator",
-      "Initiator handle",
-      "The TpHandle of the initiator of this D-Bus tube object.",
-      0, G_MAXUINT32, 0,
-      G_PARAM_CONSTRUCT_ONLY |
-      G_PARAM_READWRITE |
-      G_PARAM_STATIC_NAME |
-      G_PARAM_STATIC_NICK |
-      G_PARAM_STATIC_BLURB);
-  g_object_class_install_property (object_class, PROP_INITIATOR, param_spec);
-
-  param_spec = g_param_spec_string (
-      "service",
-      "service name",
-      "the service associated with this D-BUS tube object.",
-      "",
-      G_PARAM_CONSTRUCT_ONLY |
-      G_PARAM_READWRITE |
-      G_PARAM_STATIC_NAME |
-      G_PARAM_STATIC_NICK |
-      G_PARAM_STATIC_BLURB);
-  g_object_class_install_property (object_class, PROP_SERVICE, param_spec);
-
-  param_spec = g_param_spec_boxed (
-      "parameters",
-      "parameters GHashTable",
-      "GHashTable containing parameters of this DBUS tube object.",
-      G_TYPE_HASH_TABLE,
-      G_PARAM_CONSTRUCT_ONLY |
-      G_PARAM_READWRITE |
-      G_PARAM_STATIC_NAME |
-      G_PARAM_STATIC_NICK |
-      G_PARAM_STATIC_BLURB);
-  g_object_class_install_property (object_class, PROP_PARAMETERS, param_spec);
-
-  param_spec = g_param_spec_uint (
-      "state",
-      "Tube state",
-      "The SalutTubeState of this DBUS tube object",
-      0, G_MAXUINT32, SALUT_TUBE_STATE_REMOTE_PENDING,
-      G_PARAM_READABLE |
-      G_PARAM_STATIC_NAME |
-      G_PARAM_STATIC_NICK |
-      G_PARAM_STATIC_BLURB);
-  g_object_class_install_property (object_class, PROP_STATE, param_spec);
-
   param_spec = g_param_spec_string (
       "dbus-address",
       "D-Bus address",
@@ -818,7 +823,87 @@ salut_tube_dbus_class_init (SalutTubeDBusClass *salut_tube_dbus_class)
 }
 
 static void
-data_received_cb (GibberBytestreamIBB *ibb,
+message_received (SalutTubeDBus *tube,
+                  TpHandle sender,
+                  const char *data,
+                  size_t len)
+{
+  SalutTubeDBusPrivate *priv = SALUT_TUBE_DBUS_GET_PRIVATE (tube);
+  DBusMessage *msg;
+  DBusError error = {0,};
+  guint32 serial;
+
+  if (!priv->dbus_conn)
+    {
+      DEBUG ("no D-Bus connection");
+      return;
+    }
+
+  msg = dbus_message_demarshal (data, len, &error);
+
+  if (msg == NULL)
+    {
+      /* message was corrupted */
+      DEBUG ("received corrupted message from %d: %s: %s", sender,
+          error.name, error.message);
+      dbus_error_free (&error);
+      return;
+    }
+
+  if (priv->handle_type == TP_HANDLE_TYPE_ROOM)
+    {
+      const gchar *destination;
+      const gchar *sender_name;
+
+      destination = dbus_message_get_destination (msg);
+      /* If destination is NULL this msg is broadcasted (signals) so we don't
+       * have to check it */
+      if (destination != NULL && tp_strdiff (priv->dbus_local_name,
+            destination))
+        {
+          /* This message is not intended to this tube.
+           * Discard it. */
+          DEBUG ("message not intended to this tube (destination = %s)",
+              destination);
+          goto unref;
+        }
+
+      sender_name = g_hash_table_lookup (priv->dbus_names,
+          GUINT_TO_POINTER (sender));
+
+      if (tp_strdiff (sender_name, dbus_message_get_sender (msg)))
+        {
+          DEBUG ("invalid sender %s (expected %s for sender handle %d)",
+                 dbus_message_get_sender (msg), sender_name, sender);
+          goto unref;
+        }
+    }
+
+  /* XXX: what do do if this returns FALSE? */
+  dbus_connection_send (priv->dbus_conn, msg, &serial);
+
+unref:
+  dbus_message_unref (msg);
+}
+
+static guint32
+collect_le32 (char *str)
+{
+  unsigned char *bytes = (unsigned char *) str;
+
+  return bytes[0] | (bytes[1] << 8) | (bytes[2] << 16) | (bytes[3] << 24);
+}
+
+static guint32
+collect_be32 (char *str)
+{
+  unsigned char *bytes = (unsigned char *) str;
+
+  return (bytes[0] << 24) | (bytes[1] << 16) | (bytes[2] << 24) | bytes[3];
+}
+
+static void
+data_received_cb (GibberBytestreamIface *stream,
                   const gchar *from,
                   GString *data,
                   gpointer user_data)
@@ -828,66 +913,115 @@ data_received_cb (GibberBytestreamIBB *ibb,
   TpHandleRepoIface *contact_repo = tp_base_connection_get_handles (
       (TpBaseConnection *) priv->conn, TP_HANDLE_TYPE_CONTACT);
   TpHandle sender;
-  DBusMessage *msg;
-  DBusError error = {0,};
-  guint32 serial;
-  const gchar *sender_name;
-  const gchar *destination;
-
-  if (!priv->dbus_conn)
-    {
-      DEBUG ("no D-Bus connection");
-      return;
-    }
-
-  /* XXX: This naïvely assumes that the underlying transport always gives
-   * us complete messages. This is true for IBB, at least.
-   */
-
-  msg = dbus_message_demarshal (data->str, data->len, &error);
-  if (!msg)
-    {
-      /* message was corrupted */
-      DEBUG ("received corrupted message from %s", from);
-      return;
-    }
-
-  destination = dbus_message_get_destination (msg);
-  /* If destination is NULL this msg is broadcasted (signals) so we don't have
-   * to check it */
-  if (destination != NULL && tp_strdiff (priv->dbus_local_name, destination))
-    {
-      /* This message is not intended to this tube.
-       * Discard it. */
-      DEBUG ("message not intended to this tube (destination = %s)",
-          destination);
-      dbus_message_unref (msg);
-      return;
-    }
 
   sender = tp_handle_lookup (contact_repo, from, NULL, NULL);
   if (sender == 0)
     {
       DEBUG ("unkown sender: %s", from);
-      dbus_message_unref (msg);
       return;
     }
 
-  sender_name = g_hash_table_lookup (priv->dbus_names,
-      GUINT_TO_POINTER (sender));
-
-  if (tp_strdiff (sender_name, dbus_message_get_sender (msg)))
+  if (priv->handle_type == TP_HANDLE_TYPE_CONTACT)
     {
-      DEBUG ("invalid sender %s (expected %s for sender handle %d)",
-             dbus_message_get_sender (msg), sender_name, sender);
-      dbus_message_unref (msg);
-      return;
+      GString *buf = priv->reassembly_buffer;
+
+      g_assert (buf != NULL);
+
+      g_string_append_len (buf, data->str, data->len);
+      DEBUG ("Received %" G_GSIZE_FORMAT " bytes, so we now have %"
+          G_GSIZE_FORMAT " bytes in reassembly buffer", data->len, buf->len);
+
+      /* Each D-Bus message has a 16-byte fixed header, in which
+       *
+       * * byte 0 is 'l' (ell) or 'B' for endianness
+       * * bytes 4-7 are body length "n" in bytes in that endianness
+       * * bytes 12-15 are length "m" of param array in bytes in that
+       *   endianness
+       *
+       * followed by m + n + ((8 - (m % 8)) % 8) bytes of other content.
+       */
+
+      while (buf->len >= 16)
+        {
+          guint32 body_length, params_length, m;
+
+          /* see if we have a whole message and have already calculated
+           * how many bytes it needs */
+
+          if (priv->reassembly_bytes_needed != 0)
+            {
+              if (buf->len >= priv->reassembly_bytes_needed)
+                {
+                  DEBUG ("Received complete D-Bus message of size %"
+                      G_GINT32_FORMAT, priv->reassembly_bytes_needed);
+                  message_received (tube, sender, buf->str,
+                      priv->reassembly_bytes_needed);
+                  g_string_erase (buf, 0, priv->reassembly_bytes_needed);
+                  priv->reassembly_bytes_needed = 0;
+                }
+              else
+                {
+                  /* we'll have to wait for more data */
+                  break;
+                }
+            }
+
+          if (buf->len < 16)
+            break;
+
+          /* work out how big the next message is going to be */
+
+          if (buf->str[0] == DBUS_BIG_ENDIAN)
+            {
+              body_length = collect_be32 (buf->str + 4);
+              m = collect_be32 (buf->str + 12);
+            }
+          else if (buf->str[0] == DBUS_LITTLE_ENDIAN)
+            {
+              body_length = collect_le32 (buf->str + 4);
+              m = collect_le32 (buf->str + 12);
+            }
+          else
+            {
+              DEBUG ("D-Bus message has unknown endianness byte 0x%x, "
+                  "closing tube", (unsigned int) buf->str[0]);
+              salut_tube_iface_close (SALUT_TUBE_IFACE (tube));
+              return;
+            }
+
+          /* pad to 8-byte boundary */
+          params_length = m + ((8 - (m % 8)) % 8);
+          g_assert (params_length % 8 == 0);
+          g_assert (params_length >= m);
+          g_assert (params_length < m + 8);
+
+          priv->reassembly_bytes_needed = params_length + body_length + 16;
+
+          /* n.b.: this looks as if it could be simplified to just the third
+           * test, but that would be wrong if the addition had overflowed, so
+           * don't do that. The first and second tests are sufficient to
+           * ensure no overflow on 32-bit platforms */
+          if (body_length > DBUS_MAXIMUM_MESSAGE_LENGTH ||
+              params_length > DBUS_MAXIMUM_ARRAY_LENGTH ||
+              priv->reassembly_bytes_needed > DBUS_MAXIMUM_MESSAGE_LENGTH)
+            {
+              DEBUG ("D-Bus message is too large to be valid, closing tube");
+              salut_tube_iface_close (SALUT_TUBE_IFACE (tube));
+              return;
+            }
+
+          g_assert (priv->reassembly_bytes_needed != 0);
+          DEBUG ("We need %" G_GINT32_FORMAT " bytes for the next full "
+              "message", priv->reassembly_bytes_needed);
+        }
+    }
+  else
+    {
+      /* MUC bytestreams are message-boundary preserving, which is necessary,
+       * because we can't assume we started at the beginning */
+      g_assert (GIBBER_IS_BYTESTREAM_MUC (priv->bytestream));
+      message_received (tube, sender, data->str, data->len);
     }
-
-  /* XXX: what do do if this returns FALSE? */
-  dbus_connection_send (priv->dbus_conn, msg, &serial);
-
-  dbus_message_unref (msg);
 }
 
 SalutTubeDBus *
@@ -899,10 +1033,10 @@ salut_tube_dbus_new (SalutConnection *conn,
                      TpHandle initiator,
                      const gchar *service,
                      GHashTable *parameters,
-                     const gchar *stream_id,
-                     guint id)
+                     guint id,
+                     GibberBytestreamIface *bytestream)
 {
-  return g_object_new (SALUT_TYPE_TUBE_DBUS,
+  SalutTubeDBus *tube = g_object_new (SALUT_TYPE_TUBE_DBUS,
       "connection", conn,
       "handle", handle,
       "handle-type", handle_type,
@@ -911,9 +1045,13 @@ salut_tube_dbus_new (SalutConnection *conn,
       "initiator", initiator,
       "service", service,
       "parameters", parameters,
-      "stream-id", stream_id,
       "id", id,
       NULL);
+
+  if (bytestream != NULL)
+    g_object_set (tube, "bytestream", bytestream, NULL);
+
+  return tube;
 }
 
 /*
@@ -921,13 +1059,13 @@ salut_tube_dbus_new (SalutConnection *conn,
  *
  * Implements salut_tube_iface_accept on SalutTubeIface
  */
-static void
-salut_tube_dbus_accept (SalutTubeIface *tube)
+static gboolean
+salut_tube_dbus_accept (SalutTubeIface *tube,
+                        GError **error)
 {
   SalutTubeDBus *self = SALUT_TUBE_DBUS (tube);
   SalutTubeDBusPrivate *priv = SALUT_TUBE_DBUS_GET_PRIVATE (self);
-  GibberBytestreamIBBState state;
-  gchar *stream_init_id;
+  GibberBytestreamState state;
 
   g_assert (priv->bytestream != NULL);
 
@@ -935,55 +1073,23 @@ salut_tube_dbus_accept (SalutTubeIface *tube)
       "state", &state,
       NULL);
 
-  if (state != GIBBER_BYTESTREAM_IBB_STATE_LOCAL_PENDING)
-    return;
+  if (state != GIBBER_BYTESTREAM_STATE_LOCAL_PENDING)
+    return TRUE;
 
-  g_object_get (priv->bytestream,
-      "stream-init-id", &stream_init_id,
-      NULL);
-
-  if (stream_init_id != NULL)
+  if (priv->handle_type == TP_HANDLE_TYPE_CONTACT)
     {
-      /* Bytestream was created using a SI request so
-       * we have to accept it */
-#if 0
-      LmMessage *msg;
-      LmMessageNode *si, *tube_node;
-
-      DEBUG ("accept the SI request");
-
-      msg = salut_bytestream_ibb_make_accept_iq (priv->bytestream);
-      if (msg == NULL)
-        {
-          DEBUG ("can't create SI accept IQ. Close the bytestream");
-          gibber_bytestream_ibb_close (priv->bytestream);
-          return;
-        }
-
-      si = lm_message_node_get_child_with_namespace (msg->node, "si",
-          NS_SI);
-      g_assert (si != NULL);
-
-      tube_node = lm_message_node_add_child (si, "tube", "");
-      lm_message_node_set_attribute (tube_node, "xmlns", NS_SI_TUBES_OLD);
-
-      lm_message_node_add_child (tube_node, "dbus-name",
-          priv->dbus_local_name);
-
-      salut_bytestream_ibb_accept (priv->bytestream, msg);
-
-      lm_message_unref (msg);
-      g_free (stream_init_id);
-#endif
+      /* TODO: SI reply */
     }
   else
     {
       /* No SI so the bytestream is open */
       DEBUG ("no SI, bytestream open");
       g_object_set (priv->bytestream,
-          "state", GIBBER_BYTESTREAM_IBB_STATE_OPEN,
+          "state", GIBBER_BYTESTREAM_STATE_OPEN,
           NULL);
     }
+
+  return TRUE;
 }
 
 /*
@@ -999,7 +1105,7 @@ salut_tube_dbus_close (SalutTubeIface *tube)
 
   if (priv->bytestream != NULL)
     {
-      gibber_bytestream_ibb_close (priv->bytestream);
+      gibber_bytestream_iface_close (priv->bytestream, NULL);
     }
   else
     {
@@ -1015,10 +1121,86 @@ salut_tube_dbus_close (SalutTubeIface *tube)
 
 static void
 salut_tube_dbus_add_bytestream (SalutTubeIface *tube,
-                                GibberBytestreamIBB *bytestream)
+                                GibberBytestreamIface *bytestream)
 {
   DEBUG ("D-Bus doesn't support extra bytestream");
-  gibber_bytestream_ibb_close (bytestream);
+  gibber_bytestream_iface_close (bytestream, NULL);
+}
+
+gboolean
+salut_tube_dbus_add_name (SalutTubeDBus *self,
+                         TpHandle handle,
+                         const gchar *name)
+{
+  SalutTubeDBusPrivate *priv = SALUT_TUBE_DBUS_GET_PRIVATE (self);
+  TpHandleRepoIface *contact_repo = tp_base_connection_get_handles (
+      (TpBaseConnection *) priv->conn, TP_HANDLE_TYPE_CONTACT);
+
+  g_assert (priv->handle_type == TP_HANDLE_TYPE_ROOM);
+
+  if (g_hash_table_lookup (priv->dbus_names, GUINT_TO_POINTER (handle))
+      != NULL)
+    {
+      DEBUG ("contact %d has already announced his D-Bus name", handle);
+      return FALSE;
+    }
+
+  if (g_str_has_prefix (name, ":2."))
+    {
+      gchar *supposed_name;
+      const gchar *contact_name;
+
+      contact_name = tp_handle_inspect (contact_repo, handle);
+      supposed_name = generate_dbus_unique_name (contact_name);
+
+      if (tp_strdiff (name, supposed_name))
+        {
+          DEBUG ("contact %s announces %s as D-Bus name but it should be %s",
+              contact_name, name, supposed_name);
+          g_free (supposed_name);
+          return FALSE;
+        }
+      g_free (supposed_name);
+    }
+
+  g_hash_table_insert (priv->dbus_names, GUINT_TO_POINTER (handle),
+      g_strdup (name));
+  tp_handle_ref (contact_repo, handle);
+
+  return TRUE;
+}
+
+gboolean
+salut_tube_dbus_remove_name (SalutTubeDBus *self,
+                             TpHandle handle)
+{
+  SalutTubeDBusPrivate *priv = SALUT_TUBE_DBUS_GET_PRIVATE (self);
+  TpHandleRepoIface *contact_repo = tp_base_connection_get_handles (
+      (TpBaseConnection *) priv->conn, TP_HANDLE_TYPE_CONTACT);
+  const gchar *name;
+
+  g_assert (priv->handle_type == TP_HANDLE_TYPE_ROOM);
+
+  name = g_hash_table_lookup (priv->dbus_names, GUINT_TO_POINTER (handle));
+  if (name == NULL)
+    return FALSE;
+
+  g_hash_table_remove (priv->dbus_names, GUINT_TO_POINTER (handle));
+
+  tp_handle_unref (contact_repo, handle);
+  return TRUE;
+}
+
+gboolean
+salut_tube_dbus_handle_in_names (SalutTubeDBus *self,
+                                 TpHandle handle)
+{
+  SalutTubeDBusPrivate *priv = SALUT_TUBE_DBUS_GET_PRIVATE (self);
+
+  g_assert (priv->handle_type == TP_HANDLE_TYPE_ROOM);
+
+  return (g_hash_table_lookup (priv->dbus_names, GUINT_TO_POINTER (handle))
+      != NULL);
 }
 
 static void
